import asyncio
import os
import pprint
import signal
import subprocess
import sys
import time
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest
from bluesky.run_engine import RunEngine, TransitionError
from pytest import FixtureRequest

from ophyd_async.core import (
    FilenameProvider,
    StaticFilenameProvider,
    StaticPathProvider,
)

PANDA_RECORD = str(
    Path(__file__).parent / "unit_tests" / "fastcs" / "panda" / "db" / "panda.db"
)
INCOMPLETE_BLOCK_RECORD = str(
    Path(__file__).parent
    / "unit_tests"
    / "fastcs"
    / "panda"
    / "db"
    / "incomplete_block_panda.db"
)
INCOMPLETE_RECORD = str(
    Path(__file__).parent
    / "unit_tests"
    / "fastcs"
    / "panda"
    / "db"
    / "incomplete_panda.db"
)
EXTRA_BLOCKS_RECORD = str(
    Path(__file__).parent
    / "unit_tests"
    / "fastcs"
    / "panda"
    / "db"
    / "extra_blocks_panda.db"
)


def fixture_is_used(fixture_name, session):
    """
    Helper function to check if a fixture is used in a pytest session
    """
    for item in session.items:
        for f in item.fixturenames:
            if f == fixture_name:
                return True
    return False


def pytest_collection_modifyitems(session, config, items):
    # Raise a runtime error if docker cannot communicate to the host
    # This is needed when we want to run docker fixtures in subprocesses
    # as pytest-insubprocess doesn't report fixture errors
    if fixture_is_used("docker_composer", session):
        check_docker_sock()


# Autouse fixture that will set all EPICS networking env vars to use lo interface
# to avoid false failures caused by things like firewalls blocking EPICS traffic.
@pytest.fixture(scope="session", autouse=True)
def configure_epics_environment():
    os.environ["EPICS_CAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_AUTO_ADDR_LIST"] = "NO"
    os.environ["EPICS_CA_AUTO_BEACON_ADDR_LIST"] = "NO"

    os.environ["EPICS_PVAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_AUTO_BEACON_ADDR_LIST"] = "NO"
    os.environ["EPICS_PVA_AUTO_ADDR_LIST"] = "NO"


_ALLOWED_PYTEST_TASKS = {"async_finalizer", "async_setup", "async_teardown"}


def _error_and_kill_pending_tasks(
    loop: asyncio.AbstractEventLoop, test_name: str, test_passed: bool
) -> set[asyncio.Task]:
    """Cancels pending tasks in the event loop for a test. Raises an exception if
    the test hasn't already.

    Args:
        loop: The event loop to check for pending tasks.
        test_name: The name of the test.
        test_passed: Indicates whether the test passed.

    Returns:
        set[asyncio.Task]: The set of unfinished tasks that were cancelled.

    Raises:
        RuntimeError: If there are unfinished tasks and the test didn't fail.
    """
    unfinished_tasks = {
        task
        for task in asyncio.all_tasks(loop)
        if (coro := task.get_coro()) is not None
        and hasattr(coro, "__name__")
        and coro.__name__ not in _ALLOWED_PYTEST_TASKS
        and not task.done()
    }
    for task in unfinished_tasks:
        task.cancel()

    # We only raise an exception here if the test didn't fail anyway.
    # If it did then it makes sense that there's some tasks we need to cancel,
    # but an exception will already have been raised.
    if unfinished_tasks and test_passed:
        raise RuntimeError(
            f"Not all tasks closed during test {test_name}:\n"
            f"{pprint.pformat(unfinished_tasks, width=88)}"
        )

    return unfinished_tasks


@pytest.fixture(autouse=True, scope="function")
async def fail_test_on_unclosed_tasks(request: FixtureRequest):
    """Used on every test to ensure failure if there are pending tasks
    by the end of the test.
    """
    try:
        fail_count = request.session.testsfailed
        loop = asyncio.get_running_loop()

        loop.set_debug(True)

        request.addfinalizer(
            lambda: _error_and_kill_pending_tasks(
                loop, request.node.name, request.session.testsfailed == fail_count
            )
        )
    # Once https://github.com/bluesky/ophyd-async/issues/683
    # is finished we can remove this try, except.
    except RuntimeError as error:
        if str(error) != "no running event loop":
            raise error


@pytest.fixture(scope="function")
def RE(request: FixtureRequest):
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    RE = RunEngine({}, call_returns_result=True, loop=loop)
    fail_count = request.session.testsfailed

    def clean_event_loop():
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass

        loop.call_soon_threadsafe(loop.stop)
        RE._th.join()

        try:
            _error_and_kill_pending_tasks(
                loop, request.node.name, request.session.testsfailed == fail_count
            )
        finally:
            loop.close()

    request.addfinalizer(clean_event_loop)
    return RE


@pytest.fixture(scope="module", params=["pva"])
def panda_pva():
    processes = [
        subprocess.Popen(
            [
                sys.executable,
                "-m",
                "epicscorelibs.ioc",
                "-m",
                macros,
                "-d",
                PANDA_RECORD,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        for macros in [
            "INCLUDE_EXTRA_BLOCK=,INCLUDE_EXTRA_SIGNAL=",
            "EXCLUDE_WIDTH=#,IOC_NAME=PANDAQSRVIB",
            "EXCLUDE_PCAP=#,IOC_NAME=PANDAQSRVI",
        ]
    ]
    time.sleep(2)

    for p in processes:
        assert not p.poll(), p.stdout.read()

    yield processes

    for p in processes:
        p.terminate()
        p.communicate()


@pytest.fixture
async def normal_coroutine() -> tuple[Callable[[], Any], asyncio.Event]:
    is_running = asyncio.Event()

    async def inner_coroutine():
        is_running.set()
        await asyncio.sleep(0.01)

    return inner_coroutine, is_running


@pytest.fixture
async def failing_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)
        raise ValueError()

    return inner_coroutine


@pytest.fixture
def static_filename_provider():
    return StaticFilenameProvider("ophyd_async_tests")


@pytest.fixture
def static_path_provider_factory(tmp_path: Path):
    def create_static_dir_provider_given_fp(
        fp: FilenameProvider, directory_uri: str | None = None
    ):
        return StaticPathProvider(fp, tmp_path, directory_uri=directory_uri)

    return create_static_dir_provider_given_fp


@pytest.fixture
def static_path_provider(
    static_path_provider_factory: Callable,
    static_filename_provider: FilenameProvider,
):
    return static_path_provider_factory(static_filename_provider)


def check_docker_sock():
    """
    Check if the Docker (or compatible container engine) socket is accessible.

    This function attempts to run `docker info` to verify that the current user
    can communicate with the container engine. If the command fails, it raises
    a RuntimeError with guidance on how to fix common connection issues.
    """
    try:
        subprocess.run(
            ["docker", "info"],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError as err:
        message = """
            Cannot communicate with the container engine on the host.
            Please make sure $DOCKER_HOST points to the correct socket on the host.
            NOTE:
                For podman, $DOCKER_HOST is typically set by running
                    export DOCKER_HOST="unix://$XDG_RUNTIME_DIR/podman/podman.sock"
                Also, if you are using podman please enable the socket by running
                    systemctl --user enable podman --now"""
        raise RuntimeError(message) from err


@pytest.fixture(scope="module")
def docker_composer():
    def inner_docker_composer(
        docker_args: list[str] | None = None,
        docker_services: list[str] | str | None = None,
        ready_log_line: str | None = None,
        start_timeout: float | None = None,
        stop_timeout: float | None = None,
        wait_time: float | None = None,
    ):
        """
        Run a docker compose based service, optionally do the following:
        - wait a fixed time for the service to become ready
        - wait for the service to become ready by monitoring the STDOUT
        - run specific service(s)
        - raise for timeout
        E.g.:
            # run docker compose up and tear down after yielding
            docker_composer()
            # same as above but with additional args passed to docker
            docker_composer(docker_args=["-f", "./compose.yaml"])
            # run and wait for line in STDOUT before yielding
            docker_composer(ready_log_line="Listening on port ", start_timeout=10.0)
            # wait a fixed time for the service to become ready
            docker_composer(wait_time=1.0)
            # run specific sercices
            docker_composer(docker_services=["svc1","svc2"])
        """

        if docker_args is None:
            docker_args = []

        if docker_services is None:
            docker_services = []
        elif type(docker_services) is str:
            docker_services = [docker_services]

        # start docker compose as a background process
        process = subprocess.Popen(
            ["docker", "compose", *docker_args, "up", *docker_services],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            preexec_fn=os.setsid,  # To kill the whole group later
        )

        start_time = time.time()
        if ready_log_line is not None:
            try:
                assert process.stdout is not None
                for line in process.stdout:
                    print(line, end="")
                    if ready_log_line in line:
                        break
                    if (
                        start_timeout is not None
                        and time.time() - start_time > start_timeout
                    ):
                        raise TimeoutError(
                            f"docker compose with args {docker_args} timed out"
                        )
            except Exception:
                process.terminate()
                raise

        if wait_time is not None:
            time.sleep(wait_time)

        yield  # at this point service is expected to have started

        try:
            subprocess.run(
                ["docker", "compose", *docker_args, "down", *docker_services]
            )
        except subprocess.CalledProcessError as e:
            print(f"Failed to bring down docker services: {e}")

        # Terminate background process group
        try:
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
        except ProcessLookupError:
            pass  # Already exited

        # Close stdout to avoid ResourceWarning
        if process.stdout:
            process.stdout.close()

        # Ensure process has exited
        process.wait(timeout=stop_timeout)

    yield inner_docker_composer


@pytest.fixture(scope="module")
def ca_gateway(docker_composer):
    example_services_path = os.environ.get("EXAMPLE_SERVICES_PATH", None)
    if example_services_path is not None:  # user may start services manually
        yield from docker_composer(
            ["-f", f"{example_services_path}/compose.yaml"],
            docker_services="ca-gateway",
            ready_log_line="Running as user ",
        )


@pytest.fixture(scope="module")
def bl01t_di_cam_01(ca_gateway, docker_composer):
    example_services_path = os.environ.get("EXAMPLE_SERVICES_PATH", None)
    if example_services_path is not None:  # user may start services manually
        yield from docker_composer(
            ["-f", f"{example_services_path}/compose.yaml"],
            docker_services="bl01t-di-cam-01",
            ready_log_line="iocRun: All initialization complete",
        )



import warnings
from enum import Enum
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import yaml

from ._settings import SettingsProvider
from ._utils import ConfinedModel


def ndarray_representer(dumper: yaml.Dumper, array: npt.NDArray[Any]) -> yaml.Node:
    return dumper.represent_sequence(
        "tag:yaml.org,2002:seq", array.tolist(), flow_style=True
    )


def pydantic_model_abstraction_representer(
    dumper: yaml.Dumper, model: ConfinedModel
) -> yaml.Node:
    return dumper.represent_data(model.model_dump(mode="python"))


def enum_representer(dumper: yaml.Dumper, enum: Enum) -> yaml.Node:
    return dumper.represent_data(enum.value)


class YamlSettingsProvider(SettingsProvider):
    """For providing settings from yaml to signals."""

    def __init__(self, directory: Path | str):
        self._directory = Path(directory)

    def _file_path(self, name: str) -> Path:
        return self._directory / (name + ".yaml")

    async def store(self, name: str, data: dict[str, Any]):
        yaml.add_representer(np.ndarray, ndarray_representer, Dumper=yaml.Dumper)
        yaml.add_multi_representer(
            ConfinedModel,
            pydantic_model_abstraction_representer,
            Dumper=yaml.Dumper,
        )
        yaml.add_multi_representer(Enum, enum_representer, Dumper=yaml.Dumper)
        with open(self._file_path(name), "w") as file:
            yaml.dump(data, file)

    async def retrieve(self, name: str) -> dict[str, Any]:
        with open(self._file_path(name)) as file:
            data = yaml.full_load(file)
        if isinstance(data, list):
            warnings.warn(
                DeprecationWarning(
                    "Found old save file. Re-save your yaml settings file "
                    f"{self._file_path(name)} using "
                    "ophyd_async.plan_stubs.store_settings"
                ),
                stacklevel=2,
            )
            merge = {}
            for d in data:
                merge.update(d)
            return merge
        return data



import functools
from collections.abc import Awaitable, Callable, Mapping
from inspect import Parameter, signature
from typing import (
    Any,
    Generic,
    TypeVar,
    get_args,
    get_origin,
    is_typeddict,
)

from bluesky.protocols import Locatable

from ._derived_signal_backend import (
    DerivedSignalBackend,
    SignalTransformer,
    Transform,
    TransformT,
)
from ._device import Device
from ._signal import Signal, SignalR, SignalRW, SignalT, SignalW
from ._signal_backend import Primitive, SignalDatatypeT
from ._utils import cached_get_type_hints


class DerivedSignalFactory(Generic[TransformT]):
    """A factory that makes Derived Signals using a many-many Transform.

    :param transform_cls:
        The Transform subclass that will be filled in with parameters in order to
        transform raw to derived and derived to raw.
    :param set_derived:
        An optional async function that takes the output of
        `transform_cls.raw_to_derived` and applies it to the raw devices.
    :param raw_and_transform_devices_and_constants:
        Devices and Constants whose values will be passed as parameters
        to the `transform_cls`, and as arguments to `transform_cls.raw_to_derived`.
    """

    def __init__(
        self,
        transform_cls: type[TransformT],
        set_derived: Callable[..., Awaitable[None]] | None = None,
        **raw_and_transform_devices_and_constants,
    ):
        self._set_derived = set_derived
        _raw_and_transform_devices, _raw_and_transform_constants = (
            {
                k: v
                for k, v in raw_and_transform_devices_and_constants.items()
                if isinstance(v, Device)
            },
            {
                k: v
                for k, v in raw_and_transform_devices_and_constants.items()
                if isinstance(v, Primitive)
            },
        )

        # Check the raw and transform devices match the input arguments of the Transform
        if transform_cls is not Transform:
            # Populate expected parameters and types
            expected = {
                **{k: f.annotation for k, f in transform_cls.model_fields.items()},
                **_get_params_types_dict(transform_cls.raw_to_derived),
            }
            if empty_keys := [k for k, v in expected.items() if v == Parameter.empty]:
                raise TypeError(
                    f"{transform_cls.raw_to_derived} is missing a type "
                    f"hint for arguments: {empty_keys}"
                )

            # Populate received parameters and types
            # Use Primitive's type, Signal's datatype,
            # Locatable's datatype, or set type as None
            received = {
                **{
                    k: v.datatype if isinstance(v, Signal) else get_locatable_type(v)
                    for k, v in _raw_and_transform_devices.items()
                },
                **{k: type(v) for k, v in _raw_and_transform_constants.items()},
            }

            if expected != received:
                msg = (
                    f"Expected the following to be passed as keyword arguments "
                    f"{expected}, got {received}"
                )
                if set(expected.keys()) - set(received.keys()):
                    raise TypeError(msg)

                for k in set(expected.keys()):
                    if isinstance(expected[k], type):
                        if not issubclass(received[k], expected[k]):
                            raise TypeError(msg)
                    elif isinstance(expected[k], TypeVar):
                        bound = expected[k].__bound__
                        if isinstance(bound, type) and not issubclass(
                            received[k], bound
                        ):
                            raise TypeError(msg)
        self._set_derived_takes_dict = (
            is_typeddict(_get_first_arg_datatype(set_derived)) if set_derived else False
        )

        _raw_constants, _transform_constants = _partition_by_keys(
            _raw_and_transform_constants, set(transform_cls.model_fields)
        )

        _raw_devices, _transform_devices = _partition_by_keys(
            _raw_and_transform_devices, set(transform_cls.model_fields)
        )

        self._transformer = SignalTransformer(
            transform_cls,
            set_derived,
            self._set_derived_takes_dict,
            _raw_devices,
            _raw_constants,
            _transform_devices,
            _transform_constants,
        )

    def _make_signal(
        self,
        signal_cls: type[SignalT],
        datatype: type[SignalDatatypeT],
        name: str,
        units: str | None = None,
        precision: int | None = None,
    ) -> SignalT:
        # Check up front the raw_devices are of the right type for what the signal_cls
        # supports
        if issubclass(signal_cls, SignalR):
            self._transformer.raw_and_transform_readables  # noqa: B018
            self._transformer.raw_and_transform_subscribables  # noqa: B018
        if issubclass(signal_cls, SignalW) and not self._set_derived:
            msg = (
                f"Must define a set_derived method to support derived "
                f"{signal_cls.__name__}s"
            )
            raise ValueError(msg)
        if issubclass(signal_cls, SignalRW) and self._set_derived_takes_dict:
            self._transformer.raw_locatables  # noqa: B018
        backend = DerivedSignalBackend(
            datatype, name, self._transformer, units, precision
        )
        return signal_cls(backend)

    def derived_signal_r(
        self,
        datatype: type[SignalDatatypeT],
        name: str,
        units: str | None = None,
        precision: int | None = None,
    ) -> SignalR[SignalDatatypeT]:
        """Create a read only derived signal.

        :param datatype: The datatype of the derived signal value
        :param name:
            The name of the derived signal. Should be a key within the
            return value of `transform_cls.raw_to_derived()`.
        :param units: Engineering units for the derived signal
        :param precision: Number of digits after the decimal place to display
        """
        return self._make_signal(SignalR, datatype, name, units, precision)

    def derived_signal_w(
        self,
        datatype: type[SignalDatatypeT],
        name: str,
        units: str | None = None,
        precision: int | None = None,
    ) -> SignalW[SignalDatatypeT]:
        """Create a write only derived signal.

        :param datatype: The datatype of the derived signal value
        :param name:
            The name of the derived signal. Should be a key within the
            return value of `transform_cls.raw_to_derived()`.
        :param units: Engineering units for the derived signal
        :param precision: Number of digits after the decimal place to display
        """
        return self._make_signal(SignalW, datatype, name, units, precision)

    def derived_signal_rw(
        self,
        datatype: type[SignalDatatypeT],
        name: str,
        units: str | None = None,
        precision: int | None = None,
    ) -> SignalRW[SignalDatatypeT]:
        """Create a read-write derived signal.

        :param datatype: The datatype of the derived signal value
        :param name:
            The name of the derived signal. Should be a key within the
            return value of `transform_cls.raw_to_derived()`.
        :param units: Engineering units for the derived signal
        :param precision: Number of digits after the decimal place to display
        """
        return self._make_signal(SignalRW, datatype, name, units, precision)

    async def transform(self) -> TransformT:
        """Return an instance of `transform_cls` with all the parameters filled in."""
        return await self._transformer.get_transform()


def _get_return_datatype(func: Callable[..., SignalDatatypeT]) -> type[SignalDatatypeT]:
    args = cached_get_type_hints(func)
    if "return" not in args:
        msg = f"{func} does not have a type hint for it's return value"
        raise TypeError(msg)
    return args["return"]


def _get_first_arg_datatype(
    func: Callable[[SignalDatatypeT], Any],
) -> type[SignalDatatypeT]:
    args = _get_params_types_dict(func)
    if not args:
        msg = f"{func} does not have a type hinted argument"
        raise TypeError(msg)
    return list(args.values())[0]


def _get_params_types_dict(inspected_function: Callable) -> Mapping[str, Any]:
    sig = signature(inspected_function, eval_str=True)
    exclude_keys = {"self", "args", "kwargs", "cls"}
    return {k: v.annotation for k, v in sig.parameters.items() if k not in exclude_keys}


def _make_factory(
    raw_to_derived_func: Callable[..., SignalDatatypeT] | None = None,
    set_derived: Callable[[SignalDatatypeT], Awaitable[None]] | None = None,
    raw_devices_and_constants: dict[str, Device | Primitive] | None = None,
) -> DerivedSignalFactory:
    if raw_to_derived_func:

        class DerivedTransform(Transform):
            raw_to_derived = _dict_wrapper(raw_to_derived_func)

        return DerivedSignalFactory(
            DerivedTransform,
            set_derived=set_derived,
            **(raw_devices_and_constants or {}),
        )
    else:
        return DerivedSignalFactory(Transform, set_derived=set_derived)


def derived_signal_r(
    raw_to_derived: Callable[..., SignalDatatypeT],
    derived_units: str | None = None,
    derived_precision: int | None = None,
    **raw_devices_and_constants: Device | Primitive,
) -> SignalR[SignalDatatypeT]:
    """Create a read only derived signal.

    :param raw_to_derived:
        A function that takes the raw values as individual keyword arguments and
        returns the derived value.
    :param derived_units: Engineering units for the derived signal
    :param derived_precision: Number of digits after the decimal place to display
    :param raw_devices_and_constants:
        A dictionary of Devices and Constants to provide the values for raw_to_derived.
        The names of these arguments must match the arguments of raw_to_derived.
    """
    factory = _make_factory(
        raw_to_derived_func=raw_to_derived,
        raw_devices_and_constants=raw_devices_and_constants,
    )
    return factory.derived_signal_r(
        datatype=_get_return_datatype(raw_to_derived),
        name="value",
        units=derived_units,
        precision=derived_precision,
    )


def derived_signal_rw(
    raw_to_derived: Callable[..., SignalDatatypeT],
    set_derived: Callable[[SignalDatatypeT], Awaitable[None]],
    derived_units: str | None = None,
    derived_precision: int | None = None,
    **raw_devices_and_constants: Device | Primitive,
) -> SignalRW[SignalDatatypeT]:
    """Create a read-write derived signal.

    :param raw_to_derived:
        A function that takes the raw values as individual keyword arguments and
        returns the derived value.
    :param set_derived:
        A function that takes the derived value and sets the raw signals. It can
        either be an async function, or return an [](#AsyncStatus)
    :param derived_units: Engineering units for the derived signal
    :param derived_precision: Number of digits after the decimal place to display
    :param raw_devices_and_constants:
        A dictionary of Devices and Constants to provide the values for raw_to_derived.
        The names of these arguments must match the arguments of raw_to_derived.
    """
    raw_to_derived_datatype = _get_return_datatype(raw_to_derived)
    set_derived_arg_datatype = _get_first_arg_datatype(set_derived)
    if raw_to_derived_datatype != set_derived_arg_datatype:
        msg = (
            f"{raw_to_derived} has datatype {raw_to_derived_datatype} "
            f"!= {set_derived_arg_datatype} datatype {set_derived_arg_datatype}"
        )
        raise TypeError(msg)

    factory = _make_factory(
        raw_to_derived_func=raw_to_derived,
        set_derived=set_derived,
        raw_devices_and_constants=raw_devices_and_constants,
    )
    return factory.derived_signal_rw(
        datatype=raw_to_derived_datatype,
        name="value",
        units=derived_units,
        precision=derived_precision,
    )


def derived_signal_w(
    set_derived: Callable[[SignalDatatypeT], Awaitable[None]],
    derived_units: str | None = None,
    derived_precision: int | None = None,
) -> SignalW[SignalDatatypeT]:
    """Create a write only derived signal.

    :param set_derived:
        A function that takes the derived value and sets the raw signals. It can
        either be an async function, or return an [](#AsyncStatus)
    :param derived_units: Engineering units for the derived signal
    :param derived_precision: Number of digits after the decimal place to display
    """
    factory = _make_factory(set_derived=set_derived)
    return factory.derived_signal_w(
        datatype=_get_first_arg_datatype(set_derived),
        name="value",
        units=derived_units,
        precision=derived_precision,
    )


def get_locatable_type(obj: object) -> type | None:
    """Extract datatype from Locatable parent class.

    :param obj: Object with possible Locatable inheritance
    :return: Type hint associated with Locatable, or None if not found.
    """
    for base in getattr(obj.__class__, "__orig_bases__", []):
        if get_origin(base) is Locatable:
            args = get_args(base)
            if args:
                return args[0]
    return None


def _partition_by_keys(data: dict, keys: set) -> tuple[dict, dict]:
    group_excluded, group_included = {}, {}
    for k, v in data.items():
        if k in keys:
            group_included[k] = v
        else:
            group_excluded[k] = v
    return group_excluded, group_included


def _dict_wrapper(
    fn: Callable[..., SignalDatatypeT],
) -> Callable[..., dict[str, SignalDatatypeT]]:
    @functools.wraps(fn)
    def wrapped(self, **kwargs):
        return {"value": fn(**kwargs)}

    return wrapped



"""System tests for verifying Transport not available error when aioca/p4p missing."""

import subprocess
import sys
from pathlib import Path

import pytest

# Path to the helper script
HELPER_SCRIPT = Path(__file__).parent / "no_transport_helper.py"


@pytest.mark.parametrize(
    "block_protocols,use_protocol",
    [
        (["ca"], "ca"),
        (["pva"], "pva"),
        (["ca", "pva"], "ca"),
        (["ca", "pva"], "pva"),
    ],
)
def test_transport_not_available_error(block_protocols, use_protocol):
    flags = [f"--block-{protocol}" for protocol in block_protocols]
    pv_name = f"{use_protocol}://TEST-PV"
    result = subprocess.run(
        [sys.executable, str(HELPER_SCRIPT), *flags, pv_name],
        capture_output=True,
        text=True,
    )

    assert result.returncode != 0
    assert "NotImplementedError" in result.stderr
    assert "epics_signal_rw" in result.stderr
    assert result.stderr.endswith(
        f"Protocol {use_protocol} not available, "
        f"did you `pip install ophyd_async[{use_protocol}]`?\n"
    )



"""Used to test setting up signals for a PandA"""

import os
import re

import pytest

from ophyd_async.core import Device, DeviceVector, Signal
from ophyd_async.fastcs.core import fastcs_connector
from ophyd_async.fastcs.panda import (
    PcapBlock,
    PulseBlock,
    SeqBlock,
    SeqTable,
)


@pytest.fixture
async def panda_t():
    class CommonPandaBlocksNoData(Device):
        pcap: PcapBlock
        pulse: DeviceVector[PulseBlock]
        seq: DeviceVector[SeqBlock]

    class Panda(CommonPandaBlocksNoData):
        def __init__(self, uri: str, name: str = ""):
            super().__init__(
                name=name, connector=fastcs_connector(uri, self, "Is it ok?")
            )

    yield Panda


@pytest.mark.timeout(15.0 if os.name == "nt" else 4.0)
async def test_panda_with_missing_blocks(panda_pva, panda_t):
    panda = panda_t("PANDAQSRVI:", name="mypanda")

    with pytest.raises(
        RuntimeError,
        match=re.escape(
            "mypanda: cannot provision ['pcap'] from PANDAQSRVI:PVI: "
            "sub_devices={'pulse': 'PANDAQSRVI:PULSE:PVI', "
            "'ttlout': 'PANDAQSRVI:TTLOUT:PVI', 'seq': 'PANDAQSRVI:SEQ:PVI'}"
            "\nsignals={}"
            "\nIs it ok?"
        ),
    ):
        await panda.connect()


@pytest.mark.timeout(15.0 if os.name == "nt" else 4.1)
async def test_panda_with_extra_blocks_and_signals(panda_pva, panda_t):
    panda = panda_t("PANDAQSRV:")
    await panda.connect()
    assert panda.ttlout
    assert panda.ttlout[1]
    assert panda.ttlout[2]
    assert isinstance(panda.ttlout[1], Signal)
    assert panda.extra
    assert panda.extra[1]
    assert panda.extra[2]
    assert isinstance(panda.extra[1], Device)
    assert panda.pcap.newsignal


@pytest.mark.timeout(15.0 if os.name == "nt" else 5.1)
async def test_panda_gets_types_from_common_class(panda_pva, panda_t):
    panda = panda_t("PANDAQSRV:")
    pcap = panda.pcap
    await panda.connect()

    # The pre-initialized blocks are now filled
    assert pcap is panda.pcap

    # sub devices have the correct types
    assert isinstance(panda.pcap, PcapBlock)
    assert isinstance(panda.seq[1], SeqBlock)
    assert isinstance(panda.pulse[1], PulseBlock)

    # others are just Devices
    assert isinstance(panda.extra, DeviceVector)

    # predefined signals get set up with the correct datatype
    assert panda.pcap.active._connector.backend.datatype is bool

    # works with custom datatypes
    assert panda.seq[1].table._connector.backend.datatype is SeqTable

    # others are given the None datatype
    assert panda.pcap.newsignal._connector.backend.datatype is None


@pytest.mark.timeout(15.0 if os.name == "nt" else 4.5)
async def test_panda_block_missing_signals(panda_pva, panda_t):
    panda = panda_t("PANDAQSRVIB:")

    with pytest.raises(Exception) as exc:
        await panda.connect()
        assert (
            str(exc)
            == "PandA has a pulse block containing a width signal which has not been "
            + "retrieved by PVI."
        )



import asyncio
from typing import Annotated as A

from bluesky.protocols import Movable, Stoppable

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    SignalR,
    SignalRW,
    SignalX,
    StandardReadable,
    WatchableAsyncStatus,
    WatcherUpdate,
    observe_value,
    wait_for_value,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.tango.core import DevStateEnum, TangoDevice, TangoPolling


class TangoMover(TangoDevice, StandardReadable, Movable, Stoppable):
    """Tango moving device."""

    # Enter the name and type of the signals you want to use
    # If the server doesn't support events, the TangoPolling annotation gives
    # the parameters for ophyd to poll instead
    position: A[SignalRW[float], TangoPolling(0.1, 0.1, 0.1)]
    velocity: A[SignalRW[float], TangoPolling(0.1, 0.1, 0.1)]
    state: A[SignalR[DevStateEnum], TangoPolling(0.1)]
    # If a tango name clashes with a bluesky verb, add a trailing underscore
    stop_: SignalX

    def __init__(self, trl: str = "", name=""):
        super().__init__(trl, name=name)
        self.add_readables([self.position], Format.HINTED_SIGNAL)
        self.add_readables([self.velocity], Format.CONFIG_SIGNAL)
        self._set_success = True

    @WatchableAsyncStatus.wrap
    async def set(self, value: float, timeout: CalculatableTimeout = CALCULATE_TIMEOUT):
        self._set_success = True
        (old_position, velocity) = await asyncio.gather(
            self.position.get_value(), self.velocity.get_value()
        )
        # TODO: check whether Tango does work with negative velocity
        if timeout is CALCULATE_TIMEOUT and velocity == 0:
            msg = "Motor has zero velocity"
            raise ValueError(msg)
        else:
            timeout = abs(value - old_position) / velocity + DEFAULT_TIMEOUT

        if not (isinstance(timeout, float) or timeout is None):
            raise ValueError("Timeout must be a float or None")
        await self.position.set(value, timeout=timeout)

        move_status = AsyncStatus(
            wait_for_value(self.state, DevStateEnum.ON, timeout=timeout)
        )

        try:
            async for current_position in observe_value(
                self.position, done_status=move_status
            ):
                yield WatcherUpdate(
                    current=current_position,
                    initial=old_position,
                    target=value,
                    name=self.name,
                )
        except RuntimeError as exc:
            self._set_success = False
            raise RuntimeError("Motor was stopped") from exc
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    def stop(self, success: bool = True) -> AsyncStatus:
        self._set_success = success
        return self.stop_.trigger()



# This module adds tests designed to improve branching coverage.

# TODO: Set branch = true in [tool.coverage.run] block in
#       pyproject.toml to include these tests in coverage.

# TODO: Move to appropriate modules if they already exist.

# import pytest
from autodoc2.config import Config
from autodoc2.db import InMemoryDb
from bluesky.protocols import Locatable

from ophyd_async._docs_parser import ShortenedNamesRenderer  # noqa: PLC2701
from ophyd_async.core._derived_signal import get_locatable_type  # noqa: PLC2701

# src/ophyd_async/_docs_parser.py:10


class DummyRenderer(ShortenedNamesRenderer):
    def __init__(self):
        super().__init__(InMemoryDb(), Config())


def test_format_annotation_with_annotation():
    renderer = DummyRenderer()
    result = renderer.format_annotation("some.module.ClassName")
    assert "~some.module." in result or "ClassName" in result


def test_format_annotation_without_annotation():
    renderer = DummyRenderer()
    result = renderer.format_annotation(None)
    assert result == renderer.format_annotation(None)  # just ensure not crashing


# src/ophyd_async/core/_derived_signal.py:330


def test_get_locatable_type():
    class DummyLocatable(Locatable[int]):
        def set(self, _): ...
        def locate(self): ...

    class NonLocatable: ...

    class DummyNonLocatable(NonLocatable): ...

    obj_locatable = DummyLocatable()
    obj_nonlocatable = DummyNonLocatable()

    assert get_locatable_type(obj_locatable) is int
    assert get_locatable_type(obj_nonlocatable) is None



from xml.etree import ElementTree as ET

import pytest

from ophyd_async.core import StaticPathProvider, init_devices
from ophyd_async.epics import adcore, adsimdetector
from ophyd_async.epics.core import epics_signal_r


def test_setup_ndstats_raises_type_error(RE, static_path_provider: StaticPathProvider):
    with init_devices(mock=True):
        det = adsimdetector.SimDetector("PREFIX:", static_path_provider)
    with pytest.raises(
        AttributeError,
        match="det has no plugin named 'stats'",
    ):
        RE(adcore.setup_ndstats_sum(det))


async def test_nd_attributes_plan_stub(RE):
    async with init_devices(mock=True):
        stat = adcore.NDStatsIO("PREFIX:STATS:")
    param = adcore.NDAttributeParam(
        name="det-sum",
        param="sum",
        datatype=adcore.NDAttributeDataType.DOUBLE,
        description="Sum of det frame",
    )
    pv = adcore.NDAttributePv(
        name="Temperature",
        signal=epics_signal_r(str, "LINKAM:TEMP"),
        description="The sample temperature",
        dbrtype=adcore.NDAttributePvDbrType.DBR_FLOAT,
    )
    RE(adcore.setup_ndattributes(stat, [pv, param]))
    text = await stat.nd_attributes_file.get_value()
    xml = ET.fromstring(text)
    assert xml[0].tag == "Attribute"
    assert xml[0].attrib == {
        "name": "Temperature",
        "type": "EPICS_PV",
        "source": "LINKAM:TEMP",
        "dbrtype": "DBR_FLOAT",
        "description": "The sample temperature",
    }

    assert xml[1].tag == "Attribute"
    assert xml[1].attrib == {
        "name": "det-sum",
        "type": "PARAM",
        "source": "sum",
        "addr": "0",
        "datatype": "DOUBLE",
        "description": "Sum of det frame",
    }


@pytest.mark.parametrize("arg", [0.1, "string", 1, True, False])
async def test_nd_attributes_plan_stub_gives_correct_error(RE, arg):
    async with init_devices(mock=True):
        stat = adcore.NDStatsIO("PREFIX:STATS:")
    with pytest.raises(
        ValueError,
        match=f"Invalid type for ndattributes: {type(arg)}. "
        + "Expected NDAttributePv or NDAttributeParam.",
    ):
        RE(adcore.setup_ndattributes(stat, [arg]))



from ophyd_async.core import PathProvider, StandardDetector, soft_signal_rw
from ophyd_async.fastcs import odin
from ophyd_async.fastcs.core import fastcs_connector

from ._arm_logic import JungfrauArmLogic
from ._io import AcquisitionType, JungfrauDriverIO
from ._trigger_logic import JungfrauTriggerLogic


class JungfrauDetector(StandardDetector):
    """Ophyd-async implementation of a Jungfrau Detector."""

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix: str,
        hdf_suffix: str,
        name="",
    ):
        # Need to do this first so the bit depth signal exists for the TriggerLogic
        # once FastCS Jungfrau
        # has top level 'detector' and 'odin', after which, follow
        # EigerDetector as an example of correct structure
        drv_connector = fastcs_connector(prefix + drv_suffix)
        self.detector = JungfrauDriverIO(connector=drv_connector)
        drv_connector.create_children_from_annotations(self.detector)
        self.odin = odin.OdinIO(connector=fastcs_connector(prefix + hdf_suffix))
        self.acquisition_type = soft_signal_rw(
            AcquisitionType, AcquisitionType.STANDARD
        )
        self.add_detector_logics(
            JungfrauTriggerLogic(self.detector, self.acquisition_type),
            JungfrauArmLogic(self.detector),
            odin.OdinDataLogic(
                path_provider=path_provider,
                odin=self.odin,
                detector_bit_depth=self.detector.bit_depth,
            ),
        )
        super().__init__(name=name)



import asyncio
import logging
import re
import time
from asyncio import Event
from functools import partial
from typing import Any
from unittest.mock import ANY, AsyncMock, MagicMock, Mock, call

import numpy as np
import numpy.typing as npt
import pytest
from bluesky.protocols import Reading
from event_model import DataKey

from ophyd_async.core import (
    Array1D,
    AsyncReadable,
    Device,
    SignalR,
    SignalRW,
    SoftSignalBackend,
    StandardReadable,
    StrictEnum,
    callback_on_mock_put,
    init_devices,
    set_and_wait_for_other_value,
    set_and_wait_for_value,
    set_mock_put_proceeds,
    set_mock_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
    wait_for_value,
    walk_devices,
    walk_signal_sources,
)
from ophyd_async.core import (
    StandardReadableFormat as Format,
)
from ophyd_async.core._signal import (  # noqa: PLC2701
    _SignalCache,
)
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw
from ophyd_async.epics.core._signal import get_signal_backend_type  # noqa: PLC2701
from ophyd_async.testing import (
    ExampleEnum,
    ExampleTable,
    OneOfEverythingDevice,
    assert_configuration,
    assert_reading,
    assert_value,
    partial_reading,
)

_array_vals = {
    "int8a": np.array([-128, 127, 0, 1, 2, 3, 4], dtype=np.int8),
    "uint8a": np.array([0, 255, 0, 1, 2, 3, 4], dtype=np.uint8),
    "int16a": np.array([-32768, 32767, 0, 1, 2, 3, 4], dtype=np.int16),
    "uint16a": np.array([0, 65535, 0, 1, 2, 3, 4], dtype=np.uint16),
    "int32a": np.array([-2147483648, 2147483647, 0, 1, 2, 3, 4], dtype=np.int32),
    "uint32a": np.array([0, 4294967295, 0, 1, 2, 3, 4], dtype=np.uint32),
    "int64a": np.array(
        [-9223372036854775808, 9223372036854775807, 0, 1, 2, 3, 4],
        dtype=np.int64,
    ),
    "uint64a": np.array([0, 18446744073709551615, 0, 1, 2, 3, 4], dtype=np.uint64),
    "float32a": np.array(
        [
            -3.4028235e38,
            3.4028235e38,
            1.1754944e-38,
            1.4012985e-45,
            0.0000000e00,
            1.2340000e00,
            2.3400000e05,
            3.4499999e-06,
        ],
        dtype=np.float32,
    ),
    "float64a": np.array(
        [
            -1.79769313e308,
            1.79769313e308,
            2.22507386e-308,
            4.94065646e-324,
            0.00000000e000,
            1.23400000e000,
            2.34000000e005,
            3.45000000e-006,
        ],
        dtype=np.float64,
    ),
    "boola": np.array([False, False, True]),
}


@pytest.fixture
async def one_of_everything_device():
    device = OneOfEverythingDevice("everything-device")
    await device.connect()
    return device


def num_occurrences(substring: str, string: str) -> int:
    return len(list(re.finditer(re.escape(substring), string)))


def test_cannot_add_child_to_signal():
    signal = soft_signal_rw(str)
    with pytest.raises(
        KeyError,
        match="Cannot add Device or Signal child foo=<.*> of Signal, "
        "make a subclass of Device instead",
    ):
        signal.foo = signal


async def test_signal_connects_to_previous_backend(caplog):
    caplog.set_level(logging.DEBUG)
    signal = soft_signal_rw(int)
    mock_connect = Mock(side_effect=signal._connector.backend.connect)
    signal._connector.backend.connect = mock_connect
    await signal.connect()
    assert mock_connect.call_count == 1
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 1
    await asyncio.gather(signal.connect(), signal.connect(), signal.connect())
    assert mock_connect.call_count == 1
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 1


async def test_signal_connects_with_force_reconnect(caplog):
    caplog.set_level(logging.DEBUG)
    signal = soft_signal_rw(int)
    await signal.connect()
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 1
    await signal.connect(force_reconnect=True)
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 2


async def time_taken_by(coro) -> float:
    start = time.monotonic()
    await coro
    return time.monotonic() - start


async def test_set_and_wait_for_value_same_set_as_read():
    signal = epics_signal_rw(int, "pva://pv", name="signal")
    await signal.connect(mock=True)
    assert await signal.get_value() == 0
    set_mock_put_proceeds(signal, False)

    do_read_set = Event()
    callback_on_mock_put(signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_proceeds():
        await do_read_set.wait()
        set_mock_put_proceeds(signal, True)

    async def check_set_and_wait():
        await (await set_and_wait_for_value(signal, 1, timeout=0.1))

    assert await signal.get_value() == 0
    await asyncio.gather(wait_and_set_proceeds(), check_set_and_wait())
    assert await signal.get_value() == 1


async def test_set_and_wait_for_value_same_set_and_read_times_out():
    signal = epics_signal_rw(int, "pva://pv", name="signal")
    await signal.connect(mock=True)
    assert await signal.get_value() == 0
    callback_on_mock_put(signal, lambda v, _: v + 2)

    with pytest.raises(asyncio.TimeoutError):
        await set_and_wait_for_value(signal, 1, timeout=0.1)


async def test_set_and_wait_for_value_different_set_and_read():
    set_signal = epics_signal_rw(int, "pva://set", name="set-signal")
    match_signal = epics_signal_r(str, "pva://read", name="match-signal")
    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)

    do_read_set = Event()

    callback_on_mock_put(set_signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_read():
        await do_read_set.wait()
        set_mock_value(match_signal, "test")

    async def check_set_and_wait():
        status = await set_and_wait_for_other_value(
            set_signal,
            1,
            match_signal,
            "test",
            timeout=100,
            wait_for_set_completion=True,
        )
        assert await match_signal.get_value() == "test"
        assert status.done

    await asyncio.gather(wait_and_set_read(), check_set_and_wait())
    assert await set_signal.get_value() == 1


async def test_set_and_wait_behavior_with_wait_for_set_completion_false():
    set_signal = epics_signal_rw(int, "pva://set", name="set-signal")
    match_signal = epics_signal_r(str, "pva://read", name="match-signal")
    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)
    set_mock_put_proceeds(set_signal, False)

    do_read_set = Event()

    callback_on_mock_put(set_signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_read():
        await do_read_set.wait()
        set_mock_value(match_signal, "test")

    async def check_set_and_wait():
        status = await set_and_wait_for_other_value(
            set_signal,
            1,
            match_signal,
            "test",
            timeout=10,
            wait_for_set_completion=False,
        )
        assert not status.done
        assert await match_signal.get_value() == "test"
        set_mock_put_proceeds(set_signal, True)
        await status
        assert status.done
        assert await set_signal.get_value() == 1

    await asyncio.gather(wait_and_set_read(), check_set_and_wait())
    assert await set_signal.get_value() == 1


async def test_set_and_wait_for_value_different_set_and_read_times_out():
    set_signal = epics_signal_rw(int, "pva://set", name="set-signal")
    match_signal = epics_signal_r(str, "pva://read", name="match-signal")
    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)

    do_read_set = Event()

    callback_on_mock_put(set_signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_read():
        await do_read_set.wait()
        set_mock_value(match_signal, "not_test")

    async def check_set_and_wait():
        await (
            await set_and_wait_for_other_value(
                set_signal, 1, match_signal, "test", timeout=0.1
            )
        )

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.gather(wait_and_set_read(), check_set_and_wait())


@pytest.mark.timeout(3)
async def test_status_of_set_and_wait_for_value():
    set_signal = epics_signal_rw(int, "pva://signal")
    match_signal = epics_signal_rw(int, "pva://match_signal")

    async def set_match_signal_after_delay(value: Any, **kwargs):
        await asyncio.sleep(0.3)
        await match_signal.set(value / 2)

    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)
    callback_on_mock_put(set_signal, set_match_signal_after_delay)  # type: ignore

    status = await set_and_wait_for_value(set_signal, 2)
    assert status.done

    with pytest.raises(asyncio.TimeoutError):
        status = await set_and_wait_for_value(set_signal, 4, timeout=0.1)

    with pytest.raises(asyncio.TimeoutError):
        status = await set_and_wait_for_value(
            set_signal, 8, timeout=0.1, wait_for_set_completion=False
        )

    status = await set_and_wait_for_other_value(set_signal, 32, match_signal, 16)
    assert status.done

    status = await set_and_wait_for_other_value(
        set_signal, 30, match_signal, 15, wait_for_set_completion=False
    )
    assert not status.done
    await status

    with pytest.raises(asyncio.TimeoutError):
        status = await set_and_wait_for_other_value(
            set_signal, 30, match_signal, -1, timeout=0.5
        )


@pytest.mark.timeout(3)
async def test_callable_match_value_set_and_wait_for_value():
    set_signal = epics_signal_rw(int, "pva://signal")
    match_signal = epics_signal_rw(int, "pva://match_signal")

    async def set_match_signal_after_delay(value: Any, **kwargs):
        await asyncio.sleep(0.3)
        await match_signal.set(value / 2)

    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)
    callback_on_mock_put(set_signal, set_match_signal_after_delay)  # type: ignore

    def _equals_x(value, x):
        return value == x

    status = await set_and_wait_for_value(
        set_signal, 10, lambda val: _equals_x(val, 10)
    )
    assert status.done

    status = await set_and_wait_for_other_value(
        set_signal, 20, match_signal, lambda val: _equals_x(val, 10)
    )
    assert status.done

    with pytest.raises(asyncio.TimeoutError):
        status = await set_and_wait_for_other_value(
            set_signal, 30, match_signal, lambda val: _equals_x(val, -1), timeout=0.5
        )


async def test_given_callable_has_no_name_then_matcher_still_gives_timeout_error():
    set_signal = epics_signal_rw(int, "pva://signal")
    match_signal = epics_signal_rw(int, "pva://match_signal")

    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)

    class NoNameCallable:
        def __call__(self, val):
            return val == 20

    with pytest.raises(asyncio.TimeoutError):
        await set_and_wait_for_other_value(
            set_signal, 20, match_signal, NoNameCallable(), timeout=0.01
        )


async def test_partial_matcher_still_gives_timeout_error():
    set_signal = epics_signal_rw(int, "pva://signal")
    match_signal = epics_signal_rw(int, "pva://match_signal")

    await set_signal.connect(mock=True)
    await match_signal.connect(mock=True)

    with pytest.raises(asyncio.TimeoutError):
        await set_and_wait_for_other_value(
            set_signal, 20, match_signal, partial(lambda x, y: x == y, 20), timeout=0.01
        )


async def test_wait_for_value_with_value():
    signal = epics_signal_rw(str, read_pv="pva://signal", name="signal")
    await signal.connect(mock=True)
    await signal.set("blah")

    with pytest.raises(
        asyncio.TimeoutError,
        match="signal didn't match 'something' in 0.1s, last value 'blah'",
    ):
        await wait_for_value(signal, "something", timeout=0.1)
    assert await time_taken_by(wait_for_value(signal, "blah", timeout=2)) < 0.1
    t = asyncio.create_task(
        time_taken_by(wait_for_value(signal, "something else", timeout=2))
    )
    await asyncio.sleep(0.2)
    assert not t.done()
    set_mock_value(signal, "something else")
    assert 0.1 < await t < 1.0


async def test_wait_for_value_with_function():
    signal = epics_signal_rw(float, read_pv="pva://signal", name="signal")
    await signal.connect(mock=True)
    set_mock_value(signal, 45.8)

    def less_than_42(v):
        return v < 42

    with pytest.raises(
        asyncio.TimeoutError,
        match="signal didn't match less_than_42 in 0.1s, last value 45.8",
    ):
        await wait_for_value(signal, less_than_42, timeout=0.1)
    t = asyncio.create_task(
        time_taken_by(wait_for_value(signal, less_than_42, timeout=2))
    )
    await asyncio.sleep(0.2)
    assert not t.done()
    set_mock_value(signal, 41)
    assert 0.1 < await t < 1.0
    assert await time_taken_by(wait_for_value(signal, less_than_42, timeout=2)) < 0.1


@pytest.mark.parametrize(
    "signal_method,signal_class",
    [(soft_signal_r_and_setter, SignalR), (soft_signal_rw, SignalRW)],
)
async def test_create_soft_signal(signal_method, signal_class):
    signal_name = "TEST-PREFIX:SIGNAL"
    initial_value = "INITIAL"
    if signal_method == soft_signal_r_and_setter:
        signal, _ = signal_method(str, initial_value, signal_name)
    elif signal_method == soft_signal_rw:
        signal = signal_method(str, initial_value, signal_name)
    else:
        raise ValueError(signal_method)
    assert signal.source == f"soft://{signal_name}"
    assert isinstance(signal, signal_class)
    await signal.connect()
    assert isinstance(signal._connector.backend, SoftSignalBackend)
    assert (await signal.get_value()) == initial_value


def test_signal_r_cached():
    signal = soft_signal_r_and_setter(str, "INITIAL", "TEST-PREFIX:SIGNAL")[0]
    assert signal._cache is None
    with pytest.raises(RuntimeError, match=r".* not being monitored"):
        signal._backend_or_cache(cached=True)


class MockEnum(StrictEnum):
    GOOD = "Good"
    OK = "Ok"


class DummyReadableArray(StandardReadable):
    """A demo Readable to produce read and config signal"""

    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.int_value = epics_signal_r(int, prefix + "int")
            self.int_array = epics_signal_r(Array1D[np.int8], prefix + "Value")
            self.float_array = epics_signal_r(Array1D[np.float32], prefix + "Value")
        # Set name and signals for read() and read_configuration()
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.str_value = epics_signal_rw(str, prefix + "Value")
            self.strictEnum_value = epics_signal_rw(MockEnum, prefix + "array2")
        super().__init__(name=name)


@pytest.fixture
async def mock_readable():
    async with init_devices(mock=True):
        mock_readable = DummyReadableArray("SIM:READABLE:", name="mock_readable")
    yield mock_readable


async def test_assert_value(mock_readable: DummyReadableArray):
    set_mock_value(mock_readable.int_value, 168)
    await assert_value(mock_readable.int_value, 168)


async def test_assert_reading(mock_readable: DummyReadableArray):
    set_mock_value(mock_readable.int_value, 188)
    set_mock_value(mock_readable.int_array, np.array([1, 2, 4, 7]))
    set_mock_value(mock_readable.float_array, np.array([1.1231, -2.3, 451.15, 6.6233]))

    dummy_reading = {
        "mock_readable-int_value": Reading(
            {"alarm_severity": 0, "timestamp": ANY, "value": 188}
        ),
        "mock_readable-int_array": Reading(
            {"alarm_severity": 0, "timestamp": ANY, "value": [1, 2, 4, 7]}
        ),
        "mock_readable-float_array": Reading(
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": [1.1231, -2.3, 451.15, 6.6233],
            }
        ),
    }
    await assert_reading(mock_readable, dummy_reading)
    await assert_reading(mock_readable, dummy_reading, full_match=False)


async def test_assert_extra_expected_reading(mock_readable: DummyReadableArray):
    set_mock_value(mock_readable.int_value, 188)
    set_mock_value(mock_readable.int_array, np.array([1, 2, 4, 7]))
    set_mock_value(mock_readable.float_array, np.array([1.1231, -2.3, 451.15, 6.6233]))

    dummy_reading = {
        "mock_readable-int_value": {"value": 188},
        "mock_readable-int_array": {"value": [1, 2, 4, 7]},
        "mock_readable-float_array": {"value": [1.1231, -2.3, 451.15, 6.6233]},
        "bazinga": {"value": 0},
    }
    with pytest.raises(AssertionError, match=r"Right contains 1 more item:\n.*bazinga"):
        await assert_reading(mock_readable, dummy_reading)

    with pytest.raises(AssertionError, match=r"Right contains 1 more item:\n.*bazinga"):
        await assert_reading(mock_readable, dummy_reading, full_match=False)


async def test_assert_partial_reading(mock_readable: DummyReadableArray):
    set_mock_value(mock_readable.int_value, 188)
    set_mock_value(mock_readable.int_array, np.array([1, 2, 4, 7]))
    set_mock_value(mock_readable.float_array, np.array([1.1231, -2.3, 451.15, 6.6233]))

    dummy_reading = {
        "mock_readable-int_value": {"value": 188},
        "mock_readable-int_array": {"value": [1, 2, 4, 7]},
    }
    await assert_reading(mock_readable, dummy_reading, full_match=False)
    with pytest.raises(
        AssertionError, match=r"Left contains 1 more item:\n.*mock_readable-float_array"
    ):
        await assert_reading(mock_readable, dummy_reading)


async def test_assert_reading_optional_fields(
    one_of_everything_device: OneOfEverythingDevice,
):
    # alarm_severity of 0 and timestamp of ANY supplied if none given
    await assert_reading(
        one_of_everything_device.a_int, {"everything-device-a_int": partial_reading(1)}
    )

    with pytest.raises(AssertionError):
        await assert_reading(
            one_of_everything_device.a_int,
            {"everything-device-a_int": {"value": 1, "timestamp": -1}},
        )

    with pytest.raises(AssertionError):
        await assert_reading(
            one_of_everything_device.a_int,
            {"everything-device-a_int": {"value": 1, "alarm_severity": 1}},
        )

    await assert_reading(
        one_of_everything_device.a_int,
        {
            "everything-device-a_int": {
                "value": 1,
                "alarm_severity": 0,
                "timestamp": pytest.approx(time.time(), rel=1e-2),
            }
        },
    )


async def test_assert_configuration_everything(
    one_of_everything_device: OneOfEverythingDevice,
):
    await assert_configuration(
        one_of_everything_device,
        {
            "everything-device-a_int": partial_reading(1),
            "everything-device-a_float": partial_reading(1.234),
            "everything-device-a_str": partial_reading("test_string"),
            "everything-device-a_bool": partial_reading(True),
            "everything-device-a_enum": partial_reading("Bbb"),
            "everything-device-boola": partial_reading(_array_vals["boola"]),
            "everything-device-int8a": partial_reading(_array_vals["int8a"]),
            "everything-device-uint8a": partial_reading(_array_vals["uint8a"]),
            "everything-device-int16a": partial_reading(_array_vals["int16a"]),
            "everything-device-uint16a": partial_reading(_array_vals["uint16a"]),
            "everything-device-int32a": partial_reading(_array_vals["int32a"]),
            "everything-device-uint32a": partial_reading(_array_vals["uint32a"]),
            "everything-device-int64a": partial_reading(_array_vals["int64a"]),
            "everything-device-uint64a": partial_reading(_array_vals["uint64a"]),
            "everything-device-float32a": partial_reading(_array_vals["float32a"]),
            "everything-device-float64a": partial_reading(_array_vals["float64a"]),
            "everything-device-stra": partial_reading(["one", "two", "three"]),
            "everything-device-enuma": partial_reading([ExampleEnum.A, ExampleEnum.C]),
            "everything-device-table": partial_reading(
                ExampleTable(
                    a_bool=np.array([False, False, True, True]),
                    a_int=np.array([1, 8, -9, 32], dtype=np.int32),
                    a_float=np.array([1.8, 8.2, -6.0, 32.9887]),
                    a_str=["Hello", "World", "Foo", "Bar"],
                    a_enum=[ExampleEnum.A, ExampleEnum.B, ExampleEnum.A, ExampleEnum.C],
                ),
            ),
            "everything-device-ndarray": partial_reading(
                np.array([[1, 2, 3], [4, 5, 6]])
            ),
        },
    )


async def test_assert_reading_everything(
    one_of_everything_device: OneOfEverythingDevice,
):
    await assert_reading(one_of_everything_device, {})
    await assert_reading(
        one_of_everything_device.a_int,
        {"everything-device-a_int": partial_reading(1)},
    )
    await assert_reading(
        one_of_everything_device.a_float,
        {"everything-device-a_float": partial_reading(1.234)},
    )
    await assert_reading(
        one_of_everything_device.a_str,
        {
            "everything-device-a_str": partial_reading("test_string"),
        },
    )
    await assert_reading(
        one_of_everything_device.a_bool,
        {"everything-device-a_bool": partial_reading(True)},
    )
    await assert_reading(
        one_of_everything_device.boola,
        {
            "everything-device-boola": partial_reading(_array_vals["boola"]),
        },
    )
    await assert_reading(
        one_of_everything_device.a_enum,
        {
            "everything-device-a_enum": partial_reading(ExampleEnum.B),
        },
    )
    await assert_reading(
        one_of_everything_device.int8a,
        {
            "everything-device-int8a": partial_reading(_array_vals["int8a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.uint8a,
        {
            "everything-device-uint8a": partial_reading(_array_vals["uint8a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.int16a,
        {
            "everything-device-int16a": partial_reading(_array_vals["int16a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.uint16a,
        {
            "everything-device-uint16a": partial_reading(_array_vals["uint16a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.int32a,
        {
            "everything-device-int32a": partial_reading(_array_vals["int32a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.uint32a,
        {
            "everything-device-uint32a": partial_reading(_array_vals["uint32a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.int64a,
        {
            "everything-device-int64a": partial_reading(_array_vals["int64a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.uint64a,
        {
            "everything-device-uint64a": partial_reading(_array_vals["uint64a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.float32a,
        {
            "everything-device-float32a": partial_reading(_array_vals["float32a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.float64a,
        {
            "everything-device-float64a": partial_reading(_array_vals["float64a"]),
        },
    )
    await assert_reading(
        one_of_everything_device.stra,
        {
            "everything-device-stra": partial_reading(["one", "two", "three"]),
        },
    )
    await assert_reading(
        one_of_everything_device.enuma,
        {
            "everything-device-enuma": partial_reading([ExampleEnum.A, ExampleEnum.C]),
        },
    )
    await assert_reading(
        one_of_everything_device.table,
        {
            "everything-device-table": partial_reading(
                ExampleTable(
                    a_bool=np.array([False, False, True, True], np.bool_),
                    a_int=np.array([1, 8, -9, 32], np.int32),
                    a_float=np.array([1.8, 8.2, -6, 32.9887], np.float64),
                    a_str=["Hello", "World", "Foo", "Bar"],
                    a_enum=[ExampleEnum.A, ExampleEnum.B, ExampleEnum.A, ExampleEnum.C],
                ),
            )
        },
    )
    await assert_reading(
        one_of_everything_device.ndarray,
        {
            "everything-device-ndarray": partial_reading(
                np.array(([1, 2, 3], [4, 5, 6]))
            ),
        },
    )


async def test_assert_reading_default_metadata(
    one_of_everything_device: OneOfEverythingDevice,
):
    await assert_reading(
        one_of_everything_device.a_int,
        {"everything-device-a_int": partial_reading(1)},
    )
    await assert_reading(
        one_of_everything_device.a_int,
        {"everything-device-a_int": {"value": 1, "timestamp": ANY}},
    )
    await assert_reading(
        one_of_everything_device.a_int,
        {
            "everything-device-a_int": {
                "value": 1,
                "timestamp": pytest.approx(time.time(), rel=1e-2),
            }
        },
    )
    await assert_reading(
        one_of_everything_device.a_int,
        {
            "everything-device-a_int": {
                "value": 1,
                "timestamp": pytest.approx(time.time(), rel=1e-2),
                "alarm_severity": 0,
            }
        },
    )
    with pytest.raises(AssertionError):
        await assert_reading(
            one_of_everything_device.a_int,
            {
                "everything-device-a_int": {
                    "value": 1,
                    "timestamp": 2 * time.time(),
                }
            },
        )
    with pytest.raises(AssertionError):
        await assert_reading(
            one_of_everything_device.a_int,
            {"everything-device-a_int": partial_reading(2)},
        )
    with pytest.raises(AssertionError):
        await assert_reading(
            one_of_everything_device.a_int,
            {"everything-device-a_int": {"value": 1, "alarm_severity": 1}},
        )


@pytest.mark.parametrize(
    "int_value, int_array, float_array",
    [
        ([128, np.array([1, 2, 4, 7]), np.array([1.1231, -2.3, 451.15, 6.6233])]),
        ([188, np.array([-5, 2, 4, 7]), np.array([1.1231, -2.3, 451.15, 6.6233])]),
        ([188, np.array([1, 2, 4, 7]), np.array([1.231, -2.3, 451.15, 6.6233])]),
    ],
)
async def test_failed_assert_reading(
    mock_readable: DummyReadableArray, int_value, int_array, float_array
):
    set_mock_value(mock_readable.int_value, 188)
    set_mock_value(mock_readable.int_array, np.array([1, 2, 4, 7]))
    set_mock_value(mock_readable.float_array, np.array([1.1231, -2.3, 451.15, 6.6233]))

    dummy_reading = {
        "mock_readable-int_value": Reading(
            {"alarm_severity": 0, "timestamp": ANY, "value": int_value}
        ),
        "mock_readable-int_array": Reading(
            {"alarm_severity": 0, "timestamp": ANY, "value": int_array}
        ),
        "mock_readable-float_array": Reading(
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": float_array,
            }
        ),
    }
    with pytest.raises(AssertionError):
        await assert_reading(mock_readable, dummy_reading)


async def test_assert_reading_without_severity():
    class Thing(Device, AsyncReadable):
        async def describe(self) -> dict[str, DataKey]:
            return {}

        async def read(self) -> dict[str, Reading]:
            return {"foo": {"value": 42, "timestamp": 0.1}}

    await assert_reading(Thing(), {"foo": partial_reading(42)})


async def test_assert_configuration(mock_readable: DummyReadableArray):
    set_mock_value(mock_readable.str_value, "haha")
    set_mock_value(mock_readable.strictEnum_value, MockEnum.GOOD)
    dummy_reading = {
        "mock_readable-str_value": Reading(
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": "haha",
            }
        ),
        "mock_readable-strictEnum_value": Reading(
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": MockEnum.GOOD,
            }
        ),
    }
    await assert_configuration(mock_readable, dummy_reading)


async def test_assert_value_everything(
    one_of_everything_device: OneOfEverythingDevice,
):
    await assert_value(one_of_everything_device.a_int, 1)
    await assert_value(one_of_everything_device.a_float, 1.234)
    await assert_value(one_of_everything_device.a_str, "test_string")
    await assert_value(one_of_everything_device.a_bool, True)
    # for bools we must provide an array not a list for approx comparison to work
    await assert_value(one_of_everything_device.a_enum, ExampleEnum.B)
    await assert_value(one_of_everything_device.boola, _array_vals["boola"])
    await assert_value(
        one_of_everything_device.int8a,
        _array_vals["int8a"],
    )
    await assert_value(one_of_everything_device.uint8a, _array_vals["uint8a"])
    await assert_value(one_of_everything_device.int16a, _array_vals["int16a"])
    await assert_value(one_of_everything_device.uint16a, _array_vals["uint16a"])
    await assert_value(one_of_everything_device.int32a, _array_vals["int32a"])
    await assert_value(one_of_everything_device.uint32a, _array_vals["uint32a"])
    await assert_value(one_of_everything_device.int64a, _array_vals["int64a"])
    await assert_value(one_of_everything_device.uint64a, _array_vals["uint64a"])
    await assert_value(one_of_everything_device.float32a, _array_vals["float32a"])
    await assert_value(one_of_everything_device.float64a, _array_vals["float64a"])
    await assert_value(one_of_everything_device.stra, ["one", "two", "three"])
    await assert_value(one_of_everything_device.enuma, [ExampleEnum.A, ExampleEnum.C])
    await assert_value(
        one_of_everything_device.table,
        ExampleTable(
            a_bool=np.array([False, False, True, True], np.bool_),
            a_int=np.array([1, 8, -9, 32], np.int32),
            a_float=np.array([1.8, 8.2, -6, 32.9887], np.float64),
            a_str=["Hello", "World", "Foo", "Bar"],
            a_enum=[ExampleEnum.A, ExampleEnum.B, ExampleEnum.A, ExampleEnum.C],
        ),
    )
    await assert_value(
        one_of_everything_device.ndarray, np.array(([1, 2, 3], [4, 5, 6]))
    )


@pytest.mark.parametrize(
    "str_value, enum_value",
    [
        ("ha", MockEnum.OK),
        ("not funny", MockEnum.GOOD),
    ],
)
async def test_assert_configuraion_fail(
    mock_readable: DummyReadableArray, str_value, enum_value
):
    set_mock_value(mock_readable.str_value, "haha")
    set_mock_value(mock_readable.strictEnum_value, MockEnum.GOOD)
    dummy_reading = {
        "mock_readable-mode": Reading(
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": str_value,
            }
        ),
        "mock_readable-mode2": Reading(
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": enum_value,
            }
        ),
    }
    with pytest.raises(AssertionError):
        await assert_configuration(mock_readable, dummy_reading)


async def test_signal_get_and_set_logging(caplog):
    caplog.set_level(logging.DEBUG)
    mock_signal_rw = epics_signal_rw(int, "pva://mock_signal", name="mock_signal")
    await mock_signal_rw.connect(mock=True)
    await mock_signal_rw.set(value=0)
    assert "Putting value 0 to backend at source" in caplog.text
    assert "Successfully put value 0 to backend at source" in caplog.text
    await mock_signal_rw.get_value()
    assert "get_value() on source" in caplog.text


async def test_subscription_logs(caplog):
    caplog.set_level(logging.DEBUG)
    mock_signal_rw = epics_signal_rw(int, "pva://mock_signal", name="mock_signal")
    await mock_signal_rw.connect(mock=True)
    cbs = []
    mock_signal_rw.subscribe_reading(cbs.append)
    assert "Making subscription" in caplog.text
    mock_signal_rw.clear_sub(cbs.append)
    assert "Closing subscription on source" in caplog.text


class SomeClass:
    def __init__(self):
        self.some_attribute = "some_attribute"

    def some_function(self):
        pass


@pytest.mark.parametrize(
    "datatype,err",
    [
        (SomeClass, "Can't make converter for %s"),
        (object, "Can't make converter for %s"),
        (dict, "Can't make converter for %s"),
        (npt.NDArray[np.str_], "Expected Array1D[dtype], got %s"),
    ],
)
async def test_signal_unknown_datatype(datatype, err):
    err_str = re.escape(err % datatype)
    with pytest.raises(TypeError, match=err_str):
        await epics_signal_rw(datatype, "pva://mock_signal").connect(mock=True)
    with pytest.raises(TypeError, match=err_str):
        await epics_signal_rw(datatype, "ca://mock_signal").connect(mock=True)
    with pytest.raises(TypeError, match=err_str):
        soft_signal_rw(datatype)


async def test_soft_signal_ndarray_can_change_dtype():
    sig = soft_signal_rw(np.ndarray)
    for dtype in (np.int32, np.float64):
        await sig.set(np.arange(4, dtype=dtype))
        assert (await sig.get_value()).dtype == dtype


@pytest.fixture
async def signal_cache() -> _SignalCache[Any]:
    backend = MagicMock()
    signal = MagicMock()
    signal.source = "test_source"
    signal.log.debug = MagicMock()
    cache = _SignalCache(backend, signal)

    # Mock the _valid event to simulate it being set
    cache._valid = AsyncMock()
    cache._valid.wait = AsyncMock()

    return cache


async def test_get_reading_runtime_error(signal_cache: _SignalCache[Any]) -> None:
    with pytest.raises(RuntimeError, match="Monitor not working"):
        await asyncio.wait_for(signal_cache.get_reading(), timeout=1.0)


def test_notify_with_reading(signal_cache):
    mock_function = Mock()
    signal_cache._reading = {"value": 42}
    signal_cache._signal.name = "test_signal"
    signal_cache._notify(mock_function)
    mock_function.assert_called_once_with({"test_signal": partial_reading(42)})


async def test_notify_runtime_error(signal_cache: _SignalCache[Any]) -> None:
    function = MagicMock()

    with pytest.raises(RuntimeError, match="Monitor not working"):
        await asyncio.wait_for(
            signal_cache._notify(function),
            timeout=1.0,
        )


def test_signal_backend_throws_type_error() -> None:
    with pytest.raises(TypeError, match="Unsupported protocol: XYZ"):
        get_signal_backend_type("XYZ")  # type: ignore


async def test_remove_non_existing_listener():
    signal_rw = soft_signal_rw(int, initial_value=4)
    cbs = []
    assert signal_rw.clear_sub(cbs.append) is None


async def test_walk_devices_returns_all_devices(mock_readable: DummyReadableArray):
    """
    Test that walk_devices returns all child devices with correct dotted paths.
    """

    # Get all devices in the tree
    devices = walk_devices(mock_readable)

    assert devices == {
        "int_value": mock_readable.int_value,
        "int_array": mock_readable.int_array,
        "float_array": mock_readable.float_array,
        "str_value": mock_readable.str_value,
        "strictEnum_value": mock_readable.strictEnum_value,
    }

    # All returned objects should be Device instances
    for dev in devices.values():
        assert isinstance(dev, Device)


async def test_walk_signal_sources_returns_signal_sources(
    mock_readable: DummyReadableArray,
):
    """
    Test that walk_signal_sources returns correct mapping
    of dotted paths to Signal sources.
    """
    sources = walk_signal_sources(mock_readable)

    expected_sources = {
        "int_value": mock_readable.int_value.source,
        "int_array": mock_readable.int_array.source,
        "float_array": mock_readable.float_array.source,
        "str_value": mock_readable.str_value.source,
        "strictEnum_value": mock_readable.strictEnum_value.source,
    }

    assert sources == expected_sources


async def test_can_unsubscribe_from_subscribe_callback():
    signal = soft_signal_rw(float)
    callback = Mock()

    def unsubscribe_if_one(value):
        if value == 1.0:
            signal.clear_sub(callback)

    callback.side_effect = unsubscribe_if_one
    signal.subscribe_reading(callback)
    signal.subscribe_reading(print)
    await signal.set(1.0)
    await signal.set(2.0)
    assert callback.mock_calls == [
        call({"": {"value": 0.0, "timestamp": ANY, "alarm_severity": 0}}),
        call({"": {"value": 1.0, "timestamp": ANY, "alarm_severity": 0}}),
        call({"": {"value": 2.0, "timestamp": ANY, "alarm_severity": 0}}),
    ]



import asyncio

import numpy as np
import pytest
from pydantic import ValidationError
from scanspec.specs import Fly, Line

from ophyd_async.core import DeviceVector, init_devices, set_mock_value
from ophyd_async.epics.motor import Motor
from ophyd_async.fastcs.core import fastcs_connector
from ophyd_async.fastcs.panda import (
    CommonPandaBlocks,
    InencBlock,
    PandaPcompDirection,
    PcompInfo,
    PosOutScaleOffset,
    ScanSpecInfo,
    ScanSpecSeqTableTriggerLogic,
    SeqTable,
    SeqTableInfo,
    SeqTrigger,
    StaticPcompTriggerLogic,
    StaticSeqTableTriggerLogic,
)


@pytest.fixture
async def mock_panda():
    class Panda(CommonPandaBlocks):
        inenc: DeviceVector[InencBlock]

        def __init__(self, uri: str, name: str = ""):
            super().__init__(name=name, connector=fastcs_connector(uri, self))

    async with init_devices(mock=True):
        mock_panda = Panda("PANDAQSRV:", "mock_panda")
    set_mock_value(mock_panda.inenc[1].val_scale, 0.02)
    set_mock_value(mock_panda.inenc[1].val_offset, 0.0)
    set_mock_value(mock_panda.inenc[2].val_scale, 0.2)
    set_mock_value(mock_panda.inenc[2].val_offset, 0.0)
    assert mock_panda.name == "mock_panda"
    return mock_panda


async def test_from_inenc(mock_panda):
    panda = mock_panda
    number = 1
    pos_out_scale_offset = PosOutScaleOffset.from_inenc(panda, number)
    assert pos_out_scale_offset.name == "INENC1.VAL"
    assert await pos_out_scale_offset.scale.get_value() == 0.02
    assert await pos_out_scale_offset.offset.get_value() == 0.0


async def test_seq_table_trigger_logic(mock_panda):
    trigger_logic = StaticSeqTableTriggerLogic(mock_panda.seq[1])
    seq_table = (
        SeqTable.row(outa1=True, outa2=True)
        + SeqTable.row(outa1=False, outa2=False)
        + SeqTable.row(outa1=True, outa2=False)
        + SeqTable.row(outa1=False, outa2=True)
    )
    seq_table_info = SeqTableInfo(sequence_table=seq_table, repeats=1)

    async def set_active(value: bool):
        await asyncio.sleep(0.1)
        set_mock_value(mock_panda.seq[1].active, value)

    await trigger_logic.prepare(seq_table_info)
    await asyncio.gather(trigger_logic.kickoff(), set_active(True))
    await asyncio.gather(trigger_logic.complete(), set_active(False))


@pytest.fixture
async def sim_x_motor():
    async with init_devices(mock=True):
        sim_motor = Motor("BLxxI-MO-STAGE-01:X", name="sim_x_motor")

    yield sim_motor


@pytest.fixture
async def sim_y_motor():
    async with init_devices(mock=True):
        sim_motor = Motor("BLxxI-MO-STAGE-01:Y", name="sim_x_motor")

    yield sim_motor


async def test_seq_scanspec_trigger_logic(mock_panda, sim_x_motor, sim_y_motor) -> None:
    spec = Fly(1.0 @ (Line(sim_y_motor, 1, 2, 3) * ~Line(sim_x_motor, 1, 5, 5)))
    info = ScanSpecInfo(spec=spec, deadtime=0.1)
    trigger_logic = ScanSpecSeqTableTriggerLogic(
        mock_panda.seq[1],
        {
            sim_x_motor: PosOutScaleOffset(
                "INENC1.VAL",
                mock_panda.inenc[1].val_scale,
                mock_panda.inenc[1].val_offset,
            ),  # type: ignore
            sim_y_motor: PosOutScaleOffset(
                "INENC2.VAL",
                mock_panda.inenc[2].val_scale,
                mock_panda.inenc[2].val_offset,
            ),  # type: ignore
        },
    )
    await trigger_logic.prepare(info)
    out = await trigger_logic.seq.table.get_value()
    assert out.repeats == pytest.approx([1, 1, 1, 5, 1, 1, 1, 5, 1, 1, 1, 5])
    assert out.trigger == [
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.POSA_GT,
        SeqTrigger.IMMEDIATE,
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.POSA_LT,
        SeqTrigger.IMMEDIATE,
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.POSA_GT,
        SeqTrigger.IMMEDIATE,
    ]
    assert out.position == pytest.approx([0, 0, 25, 0, 0, 0, 275, 0, 0, 0, 25, 0])
    assert out.time1 == pytest.approx(
        [0, 0, 0, 900000, 0, 0, 0, 900000, 0, 0, 0, 900000]
    )
    assert out.time2 == pytest.approx(
        [0, 0, 0, 100000, 0, 0, 0, 100000, 0, 0, 0, 100000]
    )


async def test_seq_scanspec_trigger_logic_no_gaps(
    mock_panda, sim_x_motor, sim_y_motor
) -> None:
    spec = Fly(2.0 @ (Line(sim_y_motor, 1, 2, 3)))
    info = ScanSpecInfo(spec=spec, deadtime=0.1)
    trigger_logic = ScanSpecSeqTableTriggerLogic(
        mock_panda.seq[1],
        {
            sim_y_motor: PosOutScaleOffset(
                "INENC2.VAL",
                mock_panda.inenc[2].val_scale,
                mock_panda.inenc[2].val_offset,
            )
        },
    )
    await trigger_logic.prepare(info)
    out = await trigger_logic.seq.table.get_value()
    assert out.repeats == pytest.approx([1, 1, 1, 3])
    assert out.trigger == [
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.POSA_GT,
        SeqTrigger.IMMEDIATE,
    ]
    assert out.position == pytest.approx([0, 0, 3, 0])
    assert out.time1 == pytest.approx([0, 0, 0, 1900000])
    assert out.time2 == pytest.approx([0, 0, 0, 100000])


async def test_seq_scanspec_trigger_logic_duration_error(
    mock_panda, sim_x_motor, sim_y_motor
) -> None:
    spec = Fly(Line(sim_y_motor, 1, 2, 3) * ~Line(sim_x_motor, 1, 5, 5))
    info = ScanSpecInfo(spec=spec, deadtime=0.1)
    trigger_logic = ScanSpecSeqTableTriggerLogic(
        mock_panda.seq[1],
        {
            sim_x_motor: PosOutScaleOffset(
                "INENC1.VAL",
                mock_panda.inenc[1].val_scale,
                mock_panda.inenc[1].val_offset,
            ),
            sim_y_motor: PosOutScaleOffset(
                "INENC2.VAL",
                mock_panda.inenc[2].val_scale,
                mock_panda.inenc[2].val_offset,
            ),
        },
    )
    with pytest.raises(RuntimeError, match="Slice must have duration"):
        await trigger_logic.prepare(info)


async def test_seq_scanspec_trigger_logic_motor_not_passed(
    mock_panda, sim_x_motor, sim_y_motor
) -> None:
    spec = Fly(2.0 @ (Line(sim_y_motor, 1, 2, 3)))
    info = ScanSpecInfo(spec=spec, deadtime=0.1)
    trigger_logic = ScanSpecSeqTableTriggerLogic(
        mock_panda.seq[1],
        {
            sim_x_motor: PosOutScaleOffset(
                "INENC1.VAL",
                mock_panda.inenc[1].val_scale,
                mock_panda.inenc[1].val_offset,
            )
        },
    )
    await trigger_logic.prepare(info)
    out = await trigger_logic.seq.table.get_value()
    assert out.repeats == pytest.approx([1, 1, 3])
    assert out.trigger == [
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.IMMEDIATE,
    ]
    assert out.position == pytest.approx([0, 0, 0])
    assert out.time1 == pytest.approx([0, 0, 1900000])
    assert out.time2 == pytest.approx([0, 0, 100000])


async def test_seq_scanspec_trigger_logic_equal(
    mock_panda, sim_x_motor, sim_y_motor
) -> None:
    spec = 2.0 @ (Line(sim_x_motor, 1, 2, 3))
    info = ScanSpecInfo(spec=spec, deadtime=0.1)
    trigger_logic = ScanSpecSeqTableTriggerLogic(
        mock_panda.seq[1],
        {
            sim_x_motor: PosOutScaleOffset(
                "INENC1.VAL",
                mock_panda.inenc[1].val_scale,
                mock_panda.inenc[1].val_offset,
            )
        },
    )
    await trigger_logic.prepare(info)
    out = await trigger_logic.seq.table.get_value()
    assert out.repeats == pytest.approx([1, 1, 1, 1, 1, 1, 1, 1, 1])
    assert out.trigger == [
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.IMMEDIATE,
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.IMMEDIATE,
        SeqTrigger.BITA_0,
        SeqTrigger.BITA_1,
        SeqTrigger.IMMEDIATE,
    ]
    assert out.position == pytest.approx([0, 0, 0, 0, 0, 0, 0, 0, 0])
    assert out.time1 == pytest.approx([0, 0, 1900000, 0, 0, 1900000, 0, 0, 1900000])
    assert out.time2 == pytest.approx([0, 0, 100000, 0, 0, 100000, 0, 0, 100000])


async def test_pcomp_trigger_logic(mock_panda):
    trigger_logic = StaticPcompTriggerLogic(mock_panda.pcomp[1])
    pcomp_info = PcompInfo(
        start_postion=0,
        pulse_width=1,
        rising_edge_step=1,
        number_of_pulses=5,
        direction=PandaPcompDirection.POSITIVE,
    )

    async def set_active(value: bool):
        await asyncio.sleep(0.1)
        set_mock_value(mock_panda.pcomp[1].active, value)

    await trigger_logic.prepare(pcomp_info)
    await asyncio.gather(trigger_logic.kickoff(), set_active(True))
    await asyncio.gather(trigger_logic.complete(), set_active(False))


@pytest.mark.parametrize(
    ["kwargs", "error_msg"],
    [
        (
            {
                "sequence_table_factory": lambda: SeqTable.row(outc2=True),
                "repeats": 0,
                "prescale_as_us": -1,
            },
            "Input should be greater than or equal to 0 "
            "[type=greater_than_equal, input_value=-1, input_type=int]",
        ),
        (
            {
                "sequence_table_factory": lambda: (
                    SeqTable.row(outc2=True)
                    + SeqTable.row(outc2=False)
                    + SeqTable.row(outc2=True)
                    + SeqTable.row(outc2=False)
                ),
                "repeats": -1,
            },
            "Input should be greater than or equal to 0 "
            "[type=greater_than_equal, input_value=-1, input_type=int]",
        ),
        (
            {
                "sequence_table_factory": lambda: 1,
                "repeats": 1,
            },
            "Assertion failed, Cannot construct Table from 1 "
            "[type=assertion_error, input_value=1, input_type=int]",
        ),
    ],
)
def test_malformed_seq_table_info(kwargs, error_msg):
    with pytest.raises(ValidationError) as exc:
        SeqTableInfo(sequence_table=kwargs.pop("sequence_table_factory")(), **kwargs)
    assert error_msg in str(exc.value)


def test_malformed_trigger_in_seq_table():
    def full_seq_table(trigger):
        return SeqTable(
            repeats=np.array([1], dtype=np.uint16),
            trigger=trigger,
            position=np.array([1], dtype=np.int32),
            time1=np.array([1], dtype=np.uint32),
            outa1=np.array([1], dtype=np.bool_),
            outb1=np.array([1], dtype=np.bool_),
            outc1=np.array([1], dtype=np.bool_),
            outd1=np.array([1], dtype=np.bool_),
            oute1=np.array([1], dtype=np.bool_),
            outf1=np.array([1], dtype=np.bool_),
            time2=np.array([1], dtype=np.uint32),
            outa2=np.array([1], dtype=np.bool_),
            outb2=np.array([1], dtype=np.bool_),
            outc2=np.array([1], dtype=np.bool_),
            outd2=np.array([1], dtype=np.bool_),
            oute2=np.array([1], dtype=np.bool_),
            outf2=np.array([1], dtype=np.bool_),
        )

    for attempted_table in [
        np.array(["A"], dtype="U32"),
        np.array(["Immediate"], dtype="U32"),
        {"Immediate"},
    ]:
        with pytest.raises(ValidationError) as exc:
            full_seq_table(attempted_table)
        assert "Input should be an instance of Sequence" in str(exc)

    with pytest.raises(ValidationError) as exc:
        full_seq_table(["A"])
    assert (
        "Input should be 'Immediate', 'BITA=0', 'BITA=1', 'BITB=0', 'BITB=1', "
        "'BITC... [type=enum, input_value='A', input_type=str]"
    ) in str(exc)

    # Pydantic is able to infer type from these
    table = full_seq_table([SeqTrigger.IMMEDIATE])
    assert table.trigger == ["Immediate"] == [SeqTrigger.IMMEDIATE]
    table = full_seq_table(["Immediate"])
    assert table.trigger == ["Immediate"] == [SeqTrigger.IMMEDIATE]



import asyncio
from pathlib import Path
from unittest.mock import call

import pytest
from bluesky import RunEngine

from ophyd_async.core import (
    StandardDetector,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    callback_on_mock_put,
    init_devices,
    set_mock_value,
    soft_signal_rw,
)
from ophyd_async.fastcs.core import fastcs_connector
from ophyd_async.fastcs.odin import OdinDataLogic, OdinIO
from ophyd_async.testing import assert_has_calls

BIT_DEPTH = 16


class OdinDet(StandardDetector):
    def __init__(self, tmp_path: Path, name="", connector=None):
        path_provider = StaticPathProvider(StaticFilenameProvider("filename"), tmp_path)
        self.odin = OdinIO(connector=fastcs_connector("PREFIX:"))
        self.bit_depth = soft_signal_rw(int, BIT_DEPTH)
        self.add_detector_logics(
            OdinDataLogic(path_provider, self.odin, self.bit_depth)
        )
        super().__init__(name, connector)


@pytest.fixture
def odin_det(RE: RunEngine, tmp_path) -> OdinDet:
    with init_devices(mock=True):
        det = OdinDet(tmp_path)
    return det


async def test_describe_gives_detector_shape(odin_det: OdinDet, tmp_path):
    set_mock_value(odin_det.odin.fp.writing, True)
    set_mock_value(odin_det.odin.mw.writing, True)
    set_mock_value(odin_det.odin.fp.data_dims_1, 1024)
    set_mock_value(odin_det.odin.fp.data_dims_0, 768)
    await odin_det.prepare(TriggerInfo())
    description = await odin_det.describe()
    assert description == {
        "det": {
            "dtype": "array",
            "dtype_numpy": "<u2",
            "external": "STREAM:",
            "shape": [
                1,
                768,
                1024,
            ],
            "source": f"file://localhost/{tmp_path.as_posix().lstrip('/')}/filename.h5",
        },
    }


async def test_when_closed_then_data_capture_turned_off(odin_det: OdinDet):
    await odin_det.unstage()
    assert_has_calls(
        odin_det,
        [
            call.odin.fp.stop_writing.put(None),
            call.odin.mw.stop.put(None),
        ],
    )


async def test_wait_for_active_and_file_names_before_capture_then_wait_for_writing(
    odin_det: OdinDet, tmp_path
):
    odin: OdinIO = odin_det.odin
    ev = asyncio.Event()
    callback_on_mock_put(odin_det.odin.fp.start_writing, lambda v: ev.set())
    # Start it preparing
    status = odin_det.prepare(TriggerInfo(number_of_events=15))
    # Wait for start_writing to be called
    async with asyncio.timeout(1):
        await ev.wait()
    # Check it isn't done yet, but has the right calls
    assert not status.done
    assert_has_calls(
        odin,
        [
            call.fp.data_datatype.put("uint16"),
            call.fp.data_compression.put("BSLZ4"),
            call.fp.frames.put(0),
            call.fp.process_frames_per_block.put(1000),
            call.fp.file_path.put(str(tmp_path)),
            call.mw.directory.put(str(tmp_path)),
            call.fp.file_prefix.put("filename.h5"),
            call.mw.file_prefix.put("filename.h5"),
            call.mw.acquisition_id.put("filename.h5"),
            call.fp.start_writing.put(None),
        ],
    )
    # Set the filewriters going
    set_mock_value(odin.fp.writing, True)
    set_mock_value(odin.mw.writing, True)
    # Check we are done now, and no additional calls
    async with asyncio.timeout(1):
        await status
    assert status.done
    assert_has_calls(odin, [])


@pytest.mark.timeout(15)
async def test_hinted_fields(odin_det: OdinDet):
    set_mock_value(odin_det.odin.fp.writing, True)
    set_mock_value(odin_det.odin.mw.writing, True)
    await odin_det.prepare(TriggerInfo())
    assert odin_det.hints == {"fields": ["det"]}



from unittest.mock import call

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    OnOff,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import adaravis, adcore
from ophyd_async.testing import assert_has_calls


@pytest.fixture
async def test_adaravis(
    static_path_provider: StaticPathProvider,
) -> adaravis.AravisDetector:
    async with init_devices(mock=True):
        detector = adaravis.AravisDetector("PREFIX:", static_path_provider)
    writer = detector.get_plugin("writer", adcore.NDPluginFileIO)
    set_mock_value(writer.file_path_exists, True)
    return detector


def test_pvs_correct(test_adaravis: adaravis.AravisDetector):
    assert test_adaravis.driver.acquire.source == "mock+ca://PREFIX:cam1:Acquire_RBV"
    assert (
        test_adaravis.driver.trigger_mode.source
        == "mock+ca://PREFIX:cam1:TriggerMode_RBV"
    )


@pytest.mark.parametrize(
    "model,deadtime", [("Mako G-125", 70e-6), ("Mako G-507", 554e-6)]
)
async def test_deadtime(
    test_adaravis: adaravis.AravisDetector,
    model: str,
    deadtime: float,
):
    # Set a default model for tests that need deadtime
    set_mock_value(test_adaravis.driver.model, model)
    trigger_modes, actual_deadtime = await test_adaravis.get_trigger_deadtime()
    assert trigger_modes == {DetectorTrigger.INTERNAL, DetectorTrigger.EXTERNAL_EDGE}
    assert deadtime == actual_deadtime


async def test_prepare_external_edge(
    test_adaravis: adaravis.AravisDetector,
):
    await test_adaravis.prepare(
        TriggerInfo(
            trigger=DetectorTrigger.EXTERNAL_EDGE,
            number_of_events=5,
            livetime=0.5,
        )
    )
    assert_has_calls(
        test_adaravis.driver,
        [
            call.trigger_mode.put(OnOff.ON),
            call.trigger_source.put(adaravis.AravisTriggerSource.LINE1),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(5),
            call.acquire_time.put(0.5),
            call.acquire.put(True),
        ],
    )


async def test_prepare_internal(
    test_adaravis: adaravis.AravisDetector,
):
    await test_adaravis.prepare(TriggerInfo(number_of_events=11))
    assert_has_calls(
        test_adaravis.driver,
        [
            call.trigger_mode.put(OnOff.OFF),
            call.image_mode.put(adcore.ADImageMode.MULTIPLE),
            call.num_images.put(11),
        ],
    )



from __future__ import annotations

import logging
from unittest.mock import ANY, call

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    callback_on_mock_put,
    init_devices,
    set_mock_value,
)
from ophyd_async.fastcs.panda import (
    DatasetTable,
    HDFPanda,
    PandaCaptureMode,
    PandaHdf5DatasetType,
)
from ophyd_async.testing import assert_configuration, assert_has_calls

PANDA_DETECTOR_NAME = "mock_panda"

TABLES = [
    DatasetTable(
        name=[],
        dtype=[],
    ),
    DatasetTable(
        name=["x"],
        dtype=[PandaHdf5DatasetType.UINT_32],
    ),
    DatasetTable(
        name=[
            "x",
            "y",
            "y_min",
            "y_max",
        ],
        dtype=[
            PandaHdf5DatasetType.UINT_32,
            PandaHdf5DatasetType.FLOAT_64,
            PandaHdf5DatasetType.FLOAT_64,
            PandaHdf5DatasetType.FLOAT_64,
        ],
    ),
]


@pytest.fixture
async def hdf_panda(tmp_path) -> HDFPanda:
    path_provider = StaticPathProvider(StaticFilenameProvider("test-panda"), tmp_path)

    async with init_devices(mock=True):
        panda = HDFPanda("HDFPANDA:", path_provider=path_provider)

    def set_active(value: bool, wait: bool = True):
        set_mock_value(panda.pcap.active, value)

    callback_on_mock_put(panda.pcap.arm, set_active)
    set_mock_value(panda.data.directory_exists, True)
    set_mock_value(
        panda.data.datasets,
        DatasetTable(
            name=["x", "y"],
            dtype=[PandaHdf5DatasetType.UINT_32, PandaHdf5DatasetType.FLOAT_64],
        ),
    )

    return panda


@pytest.mark.parametrize("table", TABLES)
async def test_open_returns_correct_descriptors_and_resources(
    hdf_panda: HDFPanda, table: DatasetTable, caplog, tmp_path
):
    set_mock_value(hdf_panda.data.datasets, table)
    with caplog.at_level(logging.WARNING):
        await hdf_panda.prepare(TriggerInfo(trigger=DetectorTrigger.EXTERNAL_LEVEL))
        description = await hdf_panda.describe()

        # Check if empty datasets table leads to warning log message
        if len(table) == 0:
            assert "DATASETS table is empty!" in caplog.text

    uri = f"file://localhost/{tmp_path.as_posix().lstrip('/')}/test-panda.h5"
    for key, entry, expected_key in zip(
        description.keys(), description.values(), table.name, strict=True
    ):
        assert key == expected_key
        assert entry == {
            "source": uri,
            "shape": [1],
            "dtype": "number",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
        }

    set_mock_value(hdf_panda.data.num_captured, 7)
    docs = [doc async for doc in hdf_panda.collect_asset_docs()]
    sr = docs[: len(table)]
    sd = docs[len(table) :]

    # Expect an SR and SD for each row of the table
    assert len(sr) == len(sd) == len(table)

    for i, name in enumerate(table.name):
        assert sr[i] == (
            "stream_resource",
            {
                "data_key": name,
                "mimetype": "application/x-hdf5",
                "parameters": {
                    "chunk_shape": (1024,),
                    "dataset": f"/{name}",
                },
                "uid": ANY,
                "uri": uri,
            },
        )
        assert sd[i] == (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 7},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[i][1]["uid"],
                "uid": ANY,
            },
        )


async def test_open_sets_correct_signals(hdf_panda: HDFPanda, tmp_path):
    await hdf_panda.prepare(TriggerInfo(trigger=DetectorTrigger.EXTERNAL_LEVEL))
    assert_has_calls(
        hdf_panda,
        [
            call.data.create_directory.put(0),
            call.data.flush_period.put(0),
            call.data.hdf_directory.put(str(tmp_path)),
            call.data.hdf_file_name.put("test-panda.h5"),
            call.data.capture_mode.put(PandaCaptureMode.FOREVER),
            call.data.capture.put(True),
            call.pcap.arm.put(True),
        ],
    )


async def test_close_sets_correct_signals(hdf_panda: HDFPanda):
    await hdf_panda.unstage()
    assert_has_calls(hdf_panda.pcap, [call.arm.put(False)])
    assert_has_calls(hdf_panda.data, [call.capture.put(False)])


async def test_flyscan_documents(hdf_panda: HDFPanda, tmp_path):
    # Check there are no config sigs
    config_description = await hdf_panda.describe_configuration()
    assert config_description == {}
    await assert_configuration(hdf_panda, {})
    # Prepare for N frames
    await hdf_panda.prepare(
        TriggerInfo(
            trigger=DetectorTrigger.EXTERNAL_LEVEL,
            collections_per_event=3,
            number_of_events=5,
        )
    )
    # Check it is making 3 collections in each event
    description = await hdf_panda.describe()
    uri = f"file://localhost/{tmp_path.as_posix().lstrip('/')}/test-panda.h5"
    assert description == {
        "x": {
            "dtype": "array",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
            "shape": [3],
            "source": uri,
        },
        "y": {
            "dtype": "array",
            "dtype_numpy": "<f8",
            "external": "STREAM:",
            "shape": [3],
            "source": uri,
        },
    }
    assert hdf_panda.hints == {"fields": []}
    # Kick it off and start it completing
    await hdf_panda.kickoff()
    status = hdf_panda.complete()
    # Check it makes nothing
    docs = [doc async for doc in hdf_panda.collect_asset_docs()]
    assert docs == []
    # Make 2 and 1/3 events and check they are produced
    set_mock_value(hdf_panda.data.num_captured, 7)
    docs = [doc async for doc in hdf_panda.collect_asset_docs()]
    sr = docs[:2]
    sd = docs[2:]
    assert sr == [
        (
            "stream_resource",
            {
                "data_key": "x",
                "mimetype": "application/x-hdf5",
                "parameters": {"chunk_shape": (1024,), "dataset": "/x"},
                "uid": ANY,
                "uri": uri,
            },
        ),
        (
            "stream_resource",
            {
                "data_key": "y",
                "mimetype": "application/x-hdf5",
                "parameters": {"chunk_shape": (1024,), "dataset": "/y"},
                "uid": ANY,
                "uri": uri,
            },
        ),
    ]
    assert sd == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 2},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[0][1]["uid"],
                "uid": ANY,
            },
        ),
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 0, "stop": 2},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[1][1]["uid"],
                "uid": ANY,
            },
        ),
    ]
    assert not status.done
    # Update and check it completes
    set_mock_value(hdf_panda.data.num_captured, 15)
    set_mock_value(hdf_panda.pcap.active, False)
    docs = [doc async for doc in hdf_panda.collect_asset_docs()]
    assert docs == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 2, "stop": 5},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[0][1]["uid"],
                "uid": ANY,
            },
        ),
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 2, "stop": 5},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[1][1]["uid"],
                "uid": ANY,
            },
        ),
    ]
    assert status.done
    # Check that another kickoff without prepare is not allowed
    with pytest.raises(
        RuntimeError,
        match="Kickoff requested 15:30, but detector was only prepared up to 15",
    ):
        await hdf_panda.kickoff()
    # But preparing again is ok
    await hdf_panda.prepare(
        TriggerInfo(
            collections_per_event=3,
            number_of_events=3,
            trigger=DetectorTrigger.EXTERNAL_LEVEL,
        )
    )
    await hdf_panda.kickoff()
    # Completing adds indexes to the same resource
    status = hdf_panda.complete()
    set_mock_value(hdf_panda.data.num_captured, 24)
    set_mock_value(hdf_panda.pcap.active, False)
    docs = [doc async for doc in hdf_panda.collect_asset_docs()]
    assert docs == [
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 5, "stop": 8},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[0][1]["uid"],
                "uid": ANY,
            },
        ),
        (
            "stream_datum",
            {
                "descriptor": "",
                "indices": {"start": 5, "stop": 8},
                "seq_nums": {"start": 0, "stop": 0},
                "stream_resource": sr[1][1]["uid"],
                "uid": ANY,
            },
        ),
    ]
    assert status.done


async def test_get_indices_written(hdf_panda: HDFPanda):
    await hdf_panda.prepare(
        TriggerInfo(
            trigger=DetectorTrigger.EXTERNAL_LEVEL,
            collections_per_event=3,
            number_of_events=5,
        )
    )
    set_mock_value(hdf_panda.data.num_captured, 4)
    assert await hdf_panda.get_index() == 1  # 4 // collections_per_event
    set_mock_value(hdf_panda.data.num_captured, 12)
    assert await hdf_panda.get_index() == 4  # 12 // collections_per_event


async def test_oserror_when_hdf_dir_does_not_exist(hdf_panda: HDFPanda):
    set_mock_value(hdf_panda.data.directory_exists, False)
    with pytest.raises(OSError):
        await hdf_panda.prepare(TriggerInfo(trigger=DetectorTrigger.EXTERNAL_LEVEL))


async def test_deadtime(hdf_panda: HDFPanda):
    supported_triggers, deadtime = await hdf_panda.get_trigger_deadtime()
    assert deadtime == 8e-9
    assert supported_triggers == {DetectorTrigger.EXTERNAL_LEVEL}



from pydantic import NonNegativeInt

from ophyd_async.core import (
    Device,
    SignalR,
    SignalRW,
    SignalX,
    StrictEnum,
)


class JungfrauTriggerMode(StrictEnum):
    INTERNAL = "Internal"

    # Detector waits for external trigger to start frame series, but still
    # controls exposure time and frame period internally
    EXTERNAL = "External"


class DetectorStatus(StrictEnum):
    IDLE = "Idle"
    ERROR = "Error"
    WAITING = "Waiting"
    RUN_FINISHED = "RunFinished"
    TRANSMITTING = "Transmitting"
    RUNNING = "Running"
    STOPPED = "Stopped"


class GainMode(StrictEnum):
    DYNAMIC = "Dynamic"
    FORCE_SWITCH_G1 = "ForceSwitchG1"
    FORCE_SWITCH_G2 = "ForceSwitchG2"
    FIX_G1 = "FixG1"
    FIX_G2 = "FixG2"

    # Use with caution - this may damage the detector
    FIX_G0 = "FixG0"


class PedestalMode(StrictEnum):
    ON = "On"
    OFF = "Off"


class AcquisitionType(StrictEnum):
    STANDARD = "Standard"
    PEDESTAL = "Pedestal"


class JungfrauDriverIO(Device):
    """Contains signals for handling IO on the Jungfrau detector.

    This mirrors the intefrace provided by slsDetector/Db/slsDetector.template
    """

    exposure_time: SignalRW[float]  # in s

    # Includes deadtime
    period_between_frames: SignalRW[float]  # in s

    # Sets the delay for the beginning of the exposure time after
    # trigger input
    delay_after_trigger: SignalRW[float]  # in s

    # In internal trigger mode, this is frames per trigger. In external trigger mode,
    # this is frames per overall acquisition. In pedestal mode, this signal is not set.
    frames_per_acq: SignalRW[NonNegativeInt]

    pedestal_mode_state: SignalRW[PedestalMode]
    pedestal_mode_frames: SignalRW[NonNegativeInt]
    pedestal_mode_loops: SignalRW[NonNegativeInt]

    gain_mode: SignalRW[GainMode]

    acquisition_start: SignalX

    acquisition_stop: SignalX
    bit_depth: SignalR[int]
    trigger_mode: SignalRW[JungfrauTriggerMode]
    detector_status: SignalR[DetectorStatus]



"""The building blocks for making devices."""

from ._data_providers import (
    ReadableDataProvider,
    SignalDataProvider,
    StreamableDataProvider,
    StreamResourceDataProvider,
    StreamResourceInfo,
)
from ._datatypes import Array1D, DTypeScalar_co, Table, TableSubclass
from ._derived_signal import (
    DerivedSignalFactory,
    derived_signal_r,
    derived_signal_rw,
    derived_signal_w,
)
from ._derived_signal_backend import Transform, merge_gathered_dicts
from ._detector import (
    DetectorArmLogic,
    DetectorDataLogic,
    DetectorTrigger,
    DetectorTriggerLogic,
    StandardDetector,
    TriggerInfo,
)
from ._device import (
    Device,
    DeviceConnector,
    DeviceMock,
    DeviceVector,
    LazyMock,
    default_mock_class,
    init_devices,
)
from ._device_filler import DeviceAnnotation, DeviceFiller
from ._enums import (
    EnabledDisabled,
    EnableDisable,
    InOut,
    OnOff,
    YesNo,
)
from ._flyer import FlyerController, FlyMotorInfo, StandardFlyer
from ._log import config_ophyd_async_logging
from ._mock_signal_backend import MockSignalBackend
from ._mock_signal_utils import (
    callback_on_mock_put,
    get_mock,
    get_mock_put,
    mock_puts_blocked,
    set_mock_put_proceeds,
    set_mock_value,
    set_mock_values,
)
from ._path_providers import (
    AutoIncrementFilenameProvider,
    AutoIncrementingPathProvider,
    AutoMaxIncrementingPathProvider,
    FilenameProvider,
    PathInfo,
    PathProvider,
    StaticFilenameProvider,
    StaticPathProvider,
    UUIDFilenameProvider,
    YMDPathProvider,
)
from ._protocol import (
    AsyncConfigurable,
    AsyncLocatable,
    AsyncMovable,
    AsyncPausable,
    AsyncReadable,
    AsyncStageable,
    Watcher,
)
from ._readable import (
    ConfigSignal,
    HintedSignal,
    StandardReadable,
    StandardReadableFormat,
)
from ._settings import Settings, SettingsProvider
from ._signal import (
    Ignore,
    Signal,
    SignalConnector,
    SignalDict,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    observe_signals_value,
    observe_value,
    set_and_wait_for_other_value,
    set_and_wait_for_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
    wait_for_value,
    walk_config_signals,
    walk_devices,
    walk_rw_signals,
    walk_signal_sources,
)
from ._signal_backend import (
    Primitive,
    SignalBackend,
    SignalDatatype,
    SignalDatatypeT,
    SignalMetadata,
    make_datakey,
)
from ._soft_signal_backend import SoftSignalBackend
from ._status import AsyncStatus, WatchableAsyncStatus, completed_status
from ._utils import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    CalculatableTimeout,
    Callback,
    ConfinedModel,
    EnumTypes,
    NotConnectedError,
    Reference,
    StrictEnum,
    SubsetEnum,
    SupersetEnum,
    WatcherUpdate,
    error_if_none,
    gather_dict,
    get_dtype,
    get_enum_cls,
    get_unique,
    in_micros,
    non_zero,
    wait_for_connection,
)
from ._yaml_settings import YamlSettingsProvider


# Back compat - delete before 1.0
def __getattr__(name):
    import warnings

    renames = {
        "NotConnected": NotConnectedError,
    }
    rename = renames.get(name)
    if rename is not None:
        warnings.warn(
            DeprecationWarning(
                f"{name!r} is deprecated, use {rename.__name__!r} instead"
            ),
            stacklevel=2,
        )
        return rename
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    # Device
    "Device",
    "DeviceConnector",
    "DeviceFiller",
    "DeviceAnnotation",
    "DeviceVector",
    "init_devices",
    # Protocols
    "AsyncReadable",
    "AsyncConfigurable",
    "AsyncLocatable",
    "AsyncMovable",
    "AsyncPausable",
    "AsyncStageable",
    "Watcher",
    # Status
    "AsyncStatus",
    "WatchableAsyncStatus",
    "WatcherUpdate",
    "completed_status",
    # Signal
    "Signal",
    "SignalR",
    "SignalW",
    "SignalRW",
    "SignalX",
    "SignalBackend",
    "SignalConnector",
    # Signal Types
    "SignalDatatype",
    "SignalDatatypeT",
    "DTypeScalar_co",
    "Array1D",
    "StrictEnum",
    "SubsetEnum",
    "SupersetEnum",
    "EnumTypes",
    "Table",
    "SignalMetadata",
    "Primitive",
    # Soft signal
    "SoftSignalBackend",
    "soft_signal_r_and_setter",
    "soft_signal_rw",
    # Mock signal
    "DeviceMock",
    "LazyMock",
    "MockSignalBackend",
    "default_mock_class",
    # Mocking utilities
    "get_mock",
    "set_mock_value",
    "set_mock_values",
    "get_mock_put",
    "callback_on_mock_put",
    "mock_puts_blocked",
    "set_mock_put_proceeds",
    # Signal utilities
    "observe_value",
    "observe_signals_value",
    "wait_for_value",
    "set_and_wait_for_value",
    "set_and_wait_for_other_value",
    "walk_rw_signals",
    "walk_config_signals",
    "walk_devices",
    "walk_signal_sources",
    "SignalDict",
    # Readable
    "StandardReadable",
    "StandardReadableFormat",
    # Detector
    "DetectorTrigger",
    "TriggerInfo",
    "DetectorTriggerLogic",
    "DetectorArmLogic",
    "DetectorDataLogic",
    "StandardDetector",
    # Path
    "PathInfo",
    "PathProvider",
    "StaticPathProvider",
    "AutoIncrementingPathProvider",
    "YMDPathProvider",
    "FilenameProvider",
    "StaticFilenameProvider",
    "AutoIncrementFilenameProvider",
    "AutoMaxIncrementingPathProvider",
    "UUIDFilenameProvider",
    # Data Providers
    "ReadableDataProvider",
    "StreamableDataProvider",
    "SignalDataProvider",
    "StreamResourceInfo",
    "StreamResourceDataProvider",
    # Flyer
    "StandardFlyer",
    "FlyMotorInfo",
    "FlyerController",
    # Settings
    "Settings",
    "SettingsProvider",
    "YamlSettingsProvider",
    # Utils
    "config_ophyd_async_logging",
    "CALCULATE_TIMEOUT",
    "CalculatableTimeout",
    "DEFAULT_TIMEOUT",
    "Callback",
    "ConfinedModel",
    "NotConnectedError",
    "Reference",
    "error_if_none",
    "gather_dict",
    "get_dtype",
    "get_enum_cls",
    "get_unique",
    "in_micros",
    "make_datakey",
    "wait_for_connection",
    "Ignore",
    "non_zero",
    # Derived signal
    "derived_signal_r",
    "derived_signal_rw",
    "derived_signal_w",
    "Transform",
    "DerivedSignalFactory",
    "merge_gathered_dicts",
    # Back compat - delete before 1.0
    "ConfigSignal",
    "HintedSignal",
    # Standard enums
    "EnabledDisabled",
    "EnableDisable",
    "InOut",
    "OnOff",
    "YesNo",
    "TableSubclass",
]



import asyncio
import logging

import numpy as np

from ophyd_async.core import (
    DetectorDataLogic,
    PathProvider,
    StreamableDataProvider,
    StreamResourceDataProvider,
    StreamResourceInfo,
)

from ._block import DataBlock, PandaCaptureMode

logger = logging.getLogger("ophyd_async")


class PandaHDFDataLogic(DetectorDataLogic):
    def __init__(
        self,
        path_provider: PathProvider,
        data_block: DataBlock,
    ):
        self.path_provider = path_provider
        self.data_block = data_block

    async def prepare_unbounded(self, datakey_name: str) -> StreamableDataProvider:
        # Work out where to write
        path_info = self.path_provider(datakey_name)
        # Set create dir depth first to guarantee that callback when setting
        # directory path has correct value
        await self.data_block.create_directory.set(path_info.create_dir_depth)
        # Setup the HDF writer
        await asyncio.gather(
            self.data_block.flush_period.set(0),
            self.data_block.hdf_directory.set(str(path_info.directory_path)),
            self.data_block.hdf_file_name.set(
                f"{path_info.filename}.h5",
            ),
            self.data_block.capture_mode.set(PandaCaptureMode.FOREVER),
        )
        # Make sure that directory exists or has been created.
        if not await self.data_block.directory_exists.get_value() == 1:
            raise OSError(
                f"Directory {path_info.directory_path} does not exist or "
                "is not writable by the PandABlocks-ioc!"
            )
        # Start capturing
        await self.data_block.capture.set(True)
        # Load data from the datasets PV on the panda, update internal
        # representation of datasets that the panda will write.
        capture_table = await self.data_block.datasets.get_value()
        if len(capture_table) == 0:
            logger.warning(
                f"PandA {datakey_name} DATASETS table is empty! "
                "No stream resource docs will be generated. "
                "Make sure captured positions have their corresponding "
                "*:DATASET PV set to a scientifically relevant name."
            )
        resources = [
            StreamResourceInfo(
                data_key=dataset_name,
                shape=(),
                dtype_numpy=np.dtype(np.float64).str,
                # TODO: Update chunk size to read signal once available in IOC
                # Currently PandA IOC sets chunk size to 1024 points per chunk
                chunk_shape=(1024,),
                parameters={"dataset": f"/{dataset_name}"},
            )
            for dataset_name in capture_table.name
        ]
        return StreamResourceDataProvider(
            uri=f"{path_info.directory_uri}{path_info.filename}.h5",
            resources=resources,
            mimetype="application/x-hdf5",
            collections_written_signal=self.data_block.num_captured,
        )

    async def stop(self) -> None:
        await self.data_block.capture.set(False)



import re

import pytest

from ophyd_async.core import Device, NotConnectedError, soft_signal_rw
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.plan_stubs import ensure_connected


def test_ensure_connected(RE):
    class MyDevice(Device):
        def __init__(self, prefix: str, name=""):
            self.signal = epics_signal_rw(str, f"pva://{prefix}:SIGNAL")
            super().__init__(name=name)

    device1 = MyDevice("PREFIX1", name="device1")

    def connect():
        yield from ensure_connected(device1, mock=False, timeout=0.1)

    with pytest.raises(
        NotConnectedError,
        match="device1: NotConnectedError:\n    signal: NotConnectedError: pva://PREFIX1:SIGNAL",
    ):
        RE(connect())

    assert isinstance(device1.signal._connect_task.exception(), NotConnectedError)

    device1.signal = soft_signal_rw(str)
    RE(connect())
    assert device1.signal._connect_task.exception() is None

    device2 = MyDevice("PREFIX2", name="device2")

    def connect_with_mocking():
        assert device2.signal._mock is None
        yield from ensure_connected(device2, mock=True, timeout=0.1)
        assert device2.signal._mock is not None

    RE(connect_with_mocking())


def test_ensure_connected_fails_for_non_unique_device_names(RE):
    d1 = Device("dupe")
    d2 = Device("dupe")
    d3 = Device("ok")
    non_unique = {d1: "dupe", d2: "dupe"}
    with pytest.raises(
        ValueError,
        match=re.escape(f"Devices do not have unique names {non_unique}"),
    ):
        RE(ensure_connected(d1, d2, d3))



from pydantic import PositiveInt

from ophyd_async.core import DetectorTrigger, TriggerInfo

from ._trigger_logic import JUNGFRAU_DEADTIME_S


def create_jungfrau_external_triggering_info(
    total_triggers: PositiveInt,
    exposure_time_s: float,
) -> TriggerInfo:
    """Create safe Jungfrau TriggerInfo for external triggering.

    Uses parameters which more closely-align with Jungfrau terminology
    to create TriggerInfo. This device currently only supports one frame per trigger
    when being externally triggered.

    Args:
        total_triggers: Total external triggers expected before ending acquisition.
        exposure_time_s: How long to expose the detector for each of its frames.

    Returns:
        `TriggerInfo`
    """
    return TriggerInfo(
        number_of_events=total_triggers,
        trigger=DetectorTrigger.EXTERNAL_EDGE,
        livetime=exposure_time_s,
        deadtime=JUNGFRAU_DEADTIME_S,
    )


def create_jungfrau_internal_triggering_info(
    number_of_frames: PositiveInt, exposure_time_s: float
) -> TriggerInfo:
    """Create safe Jungfrau TriggerInfo for internal triggering.

    Uses parameters which more closely-align with Jungfrau terminology
    to create TriggerInfo.

    Args:
        number_of_frames: Total frames taken after starting acquisition.
        exposure_time_s: How long to expose the detector for each of its frames.

    Returns:
        `TriggerInfo`
    """
    return TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.INTERNAL,
        livetime=exposure_time_s,
        collections_per_event=number_of_frames,
    )


def create_jungfrau_pedestal_triggering_info(
    exposure_time_s: float,
    pedestal_frames: PositiveInt,
    pedestal_loops: PositiveInt,
):
    """Create safe Jungfrau TriggerInfo for pedestal triggering.

    Uses parameters which more closely-align with Jungfrau terminology
    to create TriggerInfo.

    When the Jungfrau is triggered in pedestal mode, it will run pedestal_frames-1
    frames in dynamic gain mode, then one frame in gain mode 1, then repeat this for
    pedelestal_loops number of times. This entire pattern is then repeated,
    but gain mode 2 is used instead of gain mode 1 for the "one frame" part.

    NOTE: To trigger the jungfrau in pedestal mode, you must first set the
    jungfrau acquisition_type signal to AcquisitionType.PEDESTAL!

    Args:
        exposure_time_s: How long to expose the detector for each of its frames.
        pedestal_frames: Number of frames taken once triggering begins
        pedestal_loops: Number of repeats of the pedestal scan before detector disarms.

    Returns:
        `TriggerInfo`
    """
    return TriggerInfo(
        collections_per_event=2 * pedestal_frames * pedestal_loops,
        trigger=DetectorTrigger.INTERNAL,
        livetime=exposure_time_s,
    )



import asyncio
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from enum import Enum
from pathlib import PureWindowsPath
from typing import Any
from xml.etree import ElementTree as ET

import numpy as np

from ophyd_async.core import (
    DetectorDataLogic,
    EnableDisable,
    PathInfo,
    PathProvider,
    SignalDataProvider,
    SignalR,
    StreamableDataProvider,
    StreamResourceDataProvider,
    StreamResourceInfo,
    set_and_wait_for_value,
)
from ophyd_async.epics.core import stop_busy_record

from ._io import (
    ADBaseColorMode,
    ADBaseDataType,
    ADBaseIO,
    ADFileWriteMode,
    NDArrayBaseIO,
    NDFileHDF5IO,
    NDPluginBaseIO,
    NDPluginFileIO,
)
from ._ndattribute import NDAttributeDataType, NDAttributePvDbrType


@dataclass
class PluginSignalDataLogic(DetectorDataLogic):
    driver: ADBaseIO
    signal: SignalR
    hinted: bool = True

    async def prepare_single(self, datakey_name: str) -> SignalDataProvider:
        # Need to wait for all the plugins to have finished before we can read
        # the plugin signal
        await self.driver.wait_for_plugins.set(True)
        return SignalDataProvider(self.signal)

    def get_hinted_fields(self, datakey_name: str) -> Sequence[str]:
        return [self.signal.name] if self.hinted else []


@dataclass
class NDArrayDescription:
    shape_signals: Sequence[SignalR[int]]
    data_type_signal: SignalR[ADBaseDataType]
    color_mode_signal: SignalR[ADBaseColorMode]


async def get_ndarray_resource_info(
    description: NDArrayDescription,
    data_key: str,
    parameters: dict[str, Any],
    frames_per_chunk: int = 1,
) -> StreamResourceInfo:
    # Grab the dimensions and datatype of the NDArray
    shape, datatype, color_mode = await asyncio.gather(
        asyncio.gather(*[sig.get_value() for sig in description.shape_signals]),
        description.data_type_signal.get_value(),
        description.color_mode_signal.get_value(),
    )
    # Remove entries in shape that are zero
    shape = [x for x in shape if x > 0]
    if datatype is ADBaseDataType.UNDEFINED:
        raise ValueError(
            f"{description.data_type_signal.source} is blank, this is not supported"
        )
    if color_mode == ADBaseColorMode.RGB1:
        shape = [3, *shape]
    elif color_mode != ADBaseColorMode.MONO:
        raise RuntimeError(
            f"Unsupported ColorMode {color_mode}! Only Mono and RGB1 are supported."
        )
    return StreamResourceInfo(
        data_key=data_key,
        shape=tuple(shape),
        chunk_shape=(frames_per_chunk, *shape),
        dtype_numpy=np.dtype(datatype.value.lower()).str,
        parameters=parameters,
    )


async def get_ndattribute_dtype_source(
    elements: Sequence[NDArrayBaseIO],
) -> dict[str, tuple[str, str]]:
    nd_attribute_xmls = await asyncio.gather(
        *[x.nd_attributes_file.get_value() for x in elements]
    )
    ndattribute_dtypes: dict[str, tuple[str, str]] = {}
    for maybe_xml in nd_attribute_xmls:
        # This is the check that ADCore does to see if it is an XML string
        # rather than a filename to parse
        if "<Attributes>" in maybe_xml:
            root = ET.fromstring(maybe_xml)
            for child in root:
                if child.attrib.get("type", "EPICS_PV") == "EPICS_PV":
                    dbrtype = child.attrib.get("dbrtype", "DBR_NATIVE")
                    if dbrtype == "DBR_NATIVE":
                        raise RuntimeError(
                            f"NDAttribute {child.attrib['name']} has dbrtype "
                            "DBR_NATIVE, which is not supported"
                        )
                    dtype_numpy = NDAttributePvDbrType[dbrtype].value
                    source = "ca://" + child.attrib["source"]
                else:
                    datatype = child.attrib.get("datatype", "INT")
                    dtype_numpy = NDAttributeDataType[datatype].value
                    source = ""
                ndattribute_dtypes[child.attrib["name"]] = (dtype_numpy, source)
    return ndattribute_dtypes


async def prepare_file_paths(
    path_info: PathInfo, file_template: str, writer: NDPluginFileIO
):
    # Set the directory creation depth first, since dir creation callback happens
    # when directory path PV is processed.
    await writer.create_directory.set(path_info.create_dir_depth)
    # When setting the path for windows based AD IOCs, areaDetector adds a '/'
    # rather than a '\\', which will cause the readback to never register the
    # same value.
    # Ensure that trailing separator is added to the directory path to avoid this.
    if isinstance(path_info.directory_path, PureWindowsPath):
        directory_path = f"{path_info.directory_path}\\"
    else:
        directory_path = f"{path_info.directory_path}/"
    await asyncio.gather(
        writer.file_path.set(directory_path),
        writer.file_name.set(path_info.filename),
        writer.file_template.set(file_template),
        writer.auto_increment.set(True),
        writer.file_number.set(0),
        writer.file_write_mode.set(ADFileWriteMode.STREAM),
    )
    # Check the path exists on the host
    if not await writer.file_path_exists.get_value():
        msg = f"Path {directory_path} doesn't exist or not writable!"
        raise FileNotFoundError(msg)
    # Overwrite num_capture to go forever
    await writer.num_capture.set(0)


@dataclass
class ADHDFDataLogic(DetectorDataLogic):
    """Data logic for AreaDetector HDF5 writer plugin.

    :param shape_signals: Signals that provide the shape of the NDArray.
    :param data_type_signal: Signal that provides the data type of the NDArray.
    :param path_provider: Callable that provides path information for file writing.
    :param driver: The AreaDetector driver instance.
    :param writer: The NDFileHDFIO plugin instance.
    :param plugins: Additional NDPluginBaseIO instances to extract NDAttributes from.
    :param datakey_suffix: Suffix to append to the data key for the main dataset
    """

    description: NDArrayDescription
    path_provider: PathProvider
    driver: ADBaseIO
    writer: NDFileHDF5IO
    plugins: Sequence[NDPluginBaseIO] = ()
    datakey_suffix: str = ""

    async def prepare_unbounded(self, datakey_name: str) -> StreamableDataProvider:
        # Work out where to write
        path_info = self.path_provider(datakey_name)
        # Determine number of frames that will be saved per HDF chunk.
        # On a fresh IOC startup, this is set to zero until the first capture,
        # so if it is zero, set it to 1.
        frames_per_chunk = await self.writer.num_frames_chunks.get_value()
        if frames_per_chunk == 0:
            frames_per_chunk = 1
            await self.writer.num_frames_chunks.set(frames_per_chunk)
        # Setup the HDF writer
        await asyncio.gather(
            self.writer.chunk_size_auto.set(True),
            self.writer.num_extra_dims.set(0),
            self.writer.lazy_open.set(True),
            self.writer.swmr_mode.set(True),
            self.writer.xml_file_name.set(""),
            self.writer.enable_callbacks.set(EnableDisable.ENABLE),
            prepare_file_paths(
                path_info=path_info, file_template="%s%s.h5", writer=self.writer
            ),
        )
        # Start capturing
        await set_and_wait_for_value(
            self.writer.capture, True, wait_for_set_completion=False
        )
        # Return a provider that reflects what we have made
        main_dataset = await get_ndarray_resource_info(
            description=self.description,
            data_key=datakey_name,
            parameters={"dataset": "/entry/data/data"},
            frames_per_chunk=frames_per_chunk,
        )
        ndattribute_dtype_sources = await get_ndattribute_dtype_source(
            (self.driver, *self.plugins)
        )
        ndattribute_datasets = [
            StreamResourceInfo(
                data_key=name,
                shape=(),
                # NDAttributes appear to always be configured with
                # this chunk size
                chunk_shape=(16384,),
                dtype_numpy=dtype_numpy,
                source=source,
                parameters={"dataset": f"/entry/instrument/NDAttributes/{name}"},
            )
            for name, (dtype_numpy, source) in ndattribute_dtype_sources.items()
        ]
        return StreamResourceDataProvider(
            uri=f"{path_info.directory_uri}{path_info.filename}.h5",
            resources=[main_dataset] + ndattribute_datasets,
            mimetype="application/x-hdf5",
            collections_written_signal=self.writer.num_captured,
            flush_signal=self.writer.flush_now,
        )

    async def stop(self) -> None:
        await stop_busy_record(self.writer.capture)

    def get_hinted_fields(self, datakey_name: str) -> Sequence[str]:
        # The main NDArray dataset is always hinted
        return [datakey_name]


@dataclass
class ADMultipartDataLogic(DetectorDataLogic):
    """Data logic for multipart AreaDetector file writers (e.g. JPEG, TIFF).

    :param shape_signals: Signals that provide the shape of the NDArray.
    :param data_type_signal: Signal that provides the data type of the NDArray.
    :param path_provider: Callable that provides path information for file writing.
    :param writer: The NDFilePluginIO instance.
    :param extension: File extension for the written files (e.g. ".jpg", ".tiff").
    :param mimetype:
        Mimetype for the written files (e.g. "multipart/related;type=image/jpeg").
    :param datakey_suffix: Suffix to append to the data key for the main dataset
    """

    description: NDArrayDescription
    path_provider: PathProvider
    writer: NDPluginFileIO
    extension: str
    mimetype: str
    datakey_suffix: str = ""

    async def prepare_unbounded(self, datakey_name: str) -> StreamableDataProvider:
        # Work out where to write
        path_info = self.path_provider(datakey_name)
        # Setup the file writer
        await prepare_file_paths(
            path_info=path_info,
            file_template="%s%s_%6.6d" + self.extension,
            writer=self.writer,
        )
        # Start capturing
        await set_and_wait_for_value(
            self.writer.capture, True, wait_for_set_completion=False
        )
        # Return a provider that reflects what we have made
        main_dataset = await get_ndarray_resource_info(
            description=self.description,
            data_key=datakey_name,
            parameters={"template": path_info.filename + "_{:06d}" + self.extension},
        )
        return StreamResourceDataProvider(
            # TODO: remove the type ignore after
            # https://github.com/bluesky/ophyd-async/issues/1186
            uri=path_info.directory_uri,  # type: ignore
            resources=[main_dataset],
            mimetype=self.mimetype,
            collections_written_signal=self.writer.num_captured,
        )

    async def stop(self) -> None:
        await stop_busy_record(self.writer.capture)

    def get_hinted_fields(self, datakey_name: str) -> Sequence[str]:
        # The main NDArray dataset is always hinted
        return [datakey_name]


class ADWriterType(Enum):
    HDF = "HDF"
    JPEG = "JPEG"
    TIFF = "TIFF"


def make_writer_data_logic(
    prefix: str,
    path_provider: PathProvider,
    writer_suffix: str | None,
    driver: ADBaseIO,
    writer_type: ADWriterType,
    plugins: Mapping[str, NDPluginBaseIO] | None = None,
) -> tuple[NDPluginFileIO, DetectorDataLogic]:
    plugins = plugins or {}
    description = NDArrayDescription(
        shape_signals=[driver.array_size_z, driver.array_size_y, driver.array_size_x],
        data_type_signal=driver.data_type,
        color_mode_signal=driver.color_mode,
    )
    match writer_type:
        case ADWriterType.HDF:
            writer = NDFileHDF5IO(f"{prefix}{writer_suffix or 'HDF1:'}")
            data_logic = ADHDFDataLogic(
                description=description,
                path_provider=path_provider,
                driver=driver,
                writer=writer,
                plugins=list(plugins.values()),
            )
        case ADWriterType.JPEG:
            writer = NDPluginFileIO(f"{prefix}{writer_suffix or 'JPEG1:'}")
            data_logic = ADMultipartDataLogic(
                description=description,
                path_provider=path_provider,
                writer=writer,
                extension=".jpg",
                mimetype="multipart/related;type=image/jpeg",
            )
        case ADWriterType.TIFF:
            writer = NDPluginFileIO(f"{prefix}{writer_suffix or 'TIFF1:'}")
            data_logic = ADMultipartDataLogic(
                description=description,
                path_provider=path_provider,
                writer=writer,
                extension=".tiff",
                mimetype="multipart/related;type=image/tiff",
            )
        case _:
            raise RuntimeError("Not implemented")
    return writer, data_logic



"""Equivalent of bluesky.protocols.Status for asynchronous tasks."""

from __future__ import annotations

import asyncio
import contextlib
import functools
import time
from asyncio import CancelledError
from collections.abc import AsyncIterator, Awaitable, Callable, Coroutine
from dataclasses import asdict, replace
from typing import Generic

from bluesky.protocols import Status

from ._device import Device
from ._protocol import Watcher
from ._utils import Callback, P, T, WatcherUpdate


class AsyncStatusBase(Status, Awaitable[None]):
    """Convert asyncio awaitable to bluesky Status interface.

    Can be used as an async context manager to automatically cancel the calling
    task when the status completes. This is useful for bounding loop execution:
    when the status completes, the calling task is cancelled, causing the loop
    to exit. If the loop completes first, the status task is automatically cancelled.
    """

    def __init__(self, awaitable: Coroutine | asyncio.Task, name: str | None = None):
        if isinstance(awaitable, asyncio.Task):
            self.task = awaitable
        else:

            async def wait_with_error_message(awaitable):
                try:
                    await awaitable
                except CancelledError as e:
                    raise CancelledError(
                        f"CancelledError while awaiting {awaitable} on {name}"
                    ) from e

            self.task = asyncio.create_task(wait_with_error_message(awaitable))
            # There is a small chance we could be cancelled before
            # wait_with_error_message starts.
            # Avoid complaints about awaitable not awaited if task is
            # pre-emptively cancelled, by ensuring it is always disposed
            self.task.add_done_callback(lambda _: awaitable.close())
        self.task.add_done_callback(self._run_callbacks)
        self._callbacks: list[Callback[Status]] = []
        self._name = name
        self._cancelled_error_ok = False

    def __await__(self):
        return self.task.__await__()

    def add_callback(self, callback: Callback[Status]):
        if self.done:
            callback(self)
        else:
            self._callbacks.append(callback)

    def _run_callbacks(self, task: asyncio.Task):
        for callback in self._callbacks:
            callback(self)

    def exception(self, timeout: float | None = 0.0) -> BaseException | None:
        """Return any exception raised by the task.

        :param timeout:
            Taken for compatibility with the Status interface, but must be 0.0 as we
            cannot wait for an async function in a sync call.
        """
        if timeout != 0.0:
            raise ValueError(
                "cannot honour any timeout other than 0 in an asynchronous function"
            )
        if self.task.done():
            try:
                return self.task.exception()
            except asyncio.CancelledError as exc:
                return exc
        return None

    @property
    def done(self) -> bool:
        return self.task.done()

    @property
    def success(self) -> bool:
        return (
            self.task.done()
            and not self.task.cancelled()
            and self.task.exception() is None
        )

    def __repr__(self) -> str:
        if self.done:
            if e := self.exception():
                status = f"errored: {repr(e)}"
            else:
                status = "done"
        else:
            status = "pending"
        device_str = f"device: {self._name}, " if self._name else ""
        return (
            f"<{type(self).__name__}, {device_str}"
            f"task: {self.task.get_coro()}, {status}>"
        )

    async def __aenter__(self):
        # Grab the calling task, the one that is doing `with status``
        calling_task = asyncio.current_task()
        if calling_task is None:
            raise RuntimeError("Can only use in a context manager inside a task")

        def _cancel_calling_task(task: asyncio.Task, calling_task=calling_task):
            # If no-one cancelled our child task, then it is expected
            # that we want to break out of the calling task with block
            # so mark that the CancelledError should be suppressed on exit
            self._cancelled_error_ok = not task.cancelled()
            calling_task.cancel()

        # When our child task is done, then cancel the calling task
        self.task.add_done_callback(_cancel_calling_task)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.task.cancel()
        # Need to await the task to suppress teardown warnings, but
        # we know it will raise CancelledError as we just cancelled it
        with contextlib.suppress(CancelledError):
            await self.task
        if exc_type is CancelledError and self._cancelled_error_ok:
            # Suppress error as we cancelled it in _cancel_calling_task
            return True
        else:
            # Raise error as we didn't cause it
            return False

    __str__ = __repr__


class AsyncStatus(AsyncStatusBase):
    """Convert an asyncio awaitable to bluesky Status interface.

    :param awaitable: The coroutine or task to await.
    :param name: The name of the device, if available.

    Can be awaited like a standard Task:

    ```python
    status = AsyncStatus(asyncio.sleep(1))
    assert not status.done
    await status  # waits for 1 second
    assert status.done
    ```

    Can also be used as a context manager to bound loop execution. When the status
    completes, the calling task is cancelled, causing loops to exit:

    ```python
    async with motor.set(target_position):
        async for value in observe_value(detector):
            process_reading(value)
            # Loop exits automatically when motor reaches position
    ```

    If the loop completes before the status, the status task is cancelled:

    ```python
    async with AsyncStatus(long_operation()):
        for i in range(3):
            await process_step(i)
        # Loop completes, long_operation() is cancelled
    ```

    Note that the body of the with statement will only break at a suspension
    point like `async for` or `await`, so body code without these suspension
    points will continue even if the status completes.
    """

    @classmethod
    def wrap(cls, f: Callable[P, Coroutine]) -> Callable[P, AsyncStatus]:
        """Wrap an async function in an AsyncStatus and return it.

        Used to make an async function conform to a bluesky protocol.

        For example:
        ```python
        class MyDevice(Device):
            @AsyncStatus.wrap
            async def trigger(self):
                await asyncio.sleep(1)
        ```
        """

        @functools.wraps(f)
        def wrap_f(*args: P.args, **kwargs: P.kwargs) -> AsyncStatus:
            if args and isinstance(args[0], Device):
                name = args[0].name
            else:
                name = None
            return cls(f(*args, **kwargs), name=name)

        return wrap_f


class WatchableAsyncStatus(AsyncStatusBase, Generic[T]):
    """Convert an asyncio async iterable to bluesky Status and Watcher interface.

    :param iterator: The async iterable to await.
    :param name: The name of the device, if available.
    """

    def __init__(
        self, iterator: AsyncIterator[WatcherUpdate[T]], name: str | None = None
    ):
        self._watchers: list[Watcher] = []
        self._start = time.monotonic()
        self._last_update: WatcherUpdate[T] | None = None
        super().__init__(self._notify_watchers_from(iterator), name)

    async def _notify_watchers_from(self, iterator: AsyncIterator[WatcherUpdate[T]]):
        async for update in iterator:
            self._last_update = (
                update
                if update.time_elapsed is not None
                else replace(update, time_elapsed=time.monotonic() - self._start)
            )
            for watcher in self._watchers:
                self._update_watcher(watcher, self._last_update)

    def _update_watcher(self, watcher: Watcher, update: WatcherUpdate[T]):
        vals = asdict(
            update, dict_factory=lambda d: {k: v for k, v in d if v is not None}
        )
        watcher(**vals)

    def watch(self, watcher: Watcher):
        """Add a watcher to the status.

        It is called:
        - immediately if there has already been an update
        - on every subsequent update
        """
        self._watchers.append(watcher)
        if self._last_update:
            self._update_watcher(watcher, self._last_update)

    @classmethod
    def wrap(
        cls,
        f: Callable[P, AsyncIterator[WatcherUpdate[T]]],
    ) -> Callable[P, WatchableAsyncStatus[T]]:
        """Wrap an AsyncIterator in a WatchableAsyncStatus.

        For example:
        ```python
        class MyDevice(Device):
            @WatchableAsyncStatus.wrap
            async def trigger(self):
                # sleep for a second, updating on progress every 0.1 seconds
                for i in range(10):
                    yield WatcherUpdate(initial=0, current=i*0.1, target=1)
                    await asyncio.sleep(0.1)
        ```
        """

        @functools.wraps(f)
        def wrap_f(*args: P.args, **kwargs: P.kwargs) -> WatchableAsyncStatus[T]:
            if args and isinstance(args[0], Device):
                name = args[0].name
            else:
                name = None
            return cls(f(*args, **kwargs), name=name)

        return wrap_f


@AsyncStatus.wrap
async def completed_status(exception: Exception | None = None):
    """Return a completed AsyncStatus.

    :param exception: If given, then raise this exception when awaited.
    """
    if exception:
        raise exception
    return None



import re
from unittest.mock import call

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    EnableDisable,
    TriggerInfo,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import adcore
from ophyd_async.testing import assert_has_calls


@pytest.fixture
async def cont_acq_detector() -> adcore.AreaDetector[adcore.ADBaseIO]:
    async with init_devices(mock=True):
        det = adcore.ContAcqDetector(prefix="PREFIX:", writer_type=None)

    set_mock_value(
        det.driver.image_mode,
        adcore.ADImageMode.CONTINUOUS,
    )
    set_mock_value(det.driver.acquire_time, 0.8)
    set_mock_value(det.driver.acquire_period, 1.0)
    set_mock_value(det.driver.acquire, True)
    return det


async def test_cont_acq_controller_invalid_trigger_mode(
    cont_acq_detector: adcore.AreaDetector[adcore.ADBaseIO],
):
    trigger_info = TriggerInfo(trigger=DetectorTrigger.EXTERNAL_EDGE)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Trigger type DetectorTrigger.EXTERNAL_EDGE not supported by 'det', "
            "supported types are: [INTERNAL]"
        ),
    ):
        await cont_acq_detector.prepare(trigger_info)


ERROR_MESSAGE = (
    "Driver must be acquiring in continuous mode to use the cont acq interface"
)


async def test_cont_acq_controller_not_in_continuous_mode(
    cont_acq_detector: adcore.AreaDetector[adcore.ADBaseIO],
):
    set_mock_value(cont_acq_detector.driver.image_mode, adcore.ADImageMode.SINGLE)
    with pytest.raises(RuntimeError, match=ERROR_MESSAGE):
        await cont_acq_detector.trigger()


async def test_cont_acq_controller_not_acquiring(
    cont_acq_detector: adcore.AreaDetector[adcore.ADBaseIO],
):
    set_mock_value(cont_acq_detector.driver.acquire, False)
    with pytest.raises(RuntimeError, match=ERROR_MESSAGE):
        await cont_acq_detector.trigger()


async def test_cont_acq_controller_invalid_exposure_time(
    cont_acq_detector: adcore.AreaDetector[adcore.ADBaseIO],
):
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Detector exposure time currently set to 0.8, but requested exposure is 1.0"
        ),
    ):
        await cont_acq_detector.prepare(TriggerInfo(livetime=1.0))


async def test_cont_acq_controller_success(
    cont_acq_detector: adcore.AreaDetector[adcore.ADBaseIO],
):
    await cont_acq_detector.stage()
    await cont_acq_detector.trigger()
    assert_has_calls(
        cont_acq_detector,
        [
            call.cb.capture.put(False),
            call.cb.enable_callbacks.put(EnableDisable.ENABLE),
            call.cb.pre_count.put(0),
            call.cb.post_count.put(1),
            call.cb.preset_trigger_count.put(1),
            call.cb.flush_on_soft_trg.put(adcore.NDCBFlushOnSoftTrgMode.ON_NEW_IMAGE),
            call.cb.capture.put(True),
        ],
    )
